# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.deprecation import deprecated
from pyomo.contrib.incidence_analysis.matching import maximum_matching
from pyomo.contrib.incidence_analysis.common.dulmage_mendelsohn import (
    # TODO: The fact that we import this function here suggests it should be
    # promoted.
    _get_projected_digraph,
)
from pyomo.common.dependencies import networkx as nx


def _get_scc_dag_of_projection(graph, top_nodes, matching):
    """Return the DAG of strongly connected components of a bipartite graph,
    projected with respect to a perfect matching

    This data structure can be used, for instance, to identify the minimal
    subsystem of constraints and variables necessary to solve a given variable
    or constraint.

    """
    nxc = nx.algorithms.components
    # _get_projected_digraph treats matched edges as "in-edges", so we
    # reverse the direction of edges here.
    dg = _get_projected_digraph(graph, matching, top_nodes).reverse()

    scc_list = list(nxc.strongly_connected_components(dg))
    n_scc = len(scc_list)
    node_scc_map = {n: idx for idx, scc in enumerate(scc_list) for n in scc}

    # Now we need to put the SCCs in the right order. We do this by performing
    # a topological sort on the DAG of SCCs.
    dag = nx.DiGraph()
    dag.add_nodes_from(range(n_scc))
    for n in dg.nodes:
        source_scc = node_scc_map[n]
        for neighbor in dg[n]:
            target_scc = node_scc_map[neighbor]
            if target_scc != source_scc:
                dag.add_edge(source_scc, target_scc)

    # Note that the matching is required to fully interpret scc_list (as it
    # only contains the "top nodes")
    return scc_list, dag


def get_scc_of_projection(graph, top_nodes, matching=None):
    """Return the topologically ordered strongly connected components of a
    bipartite graph, projected with respect to a perfect matching

    The provided undirected bipartite graph is projected into a directed graph
    on the set of "top nodes" by treating "matched edges" as out-edges and
    "unmatched edges" as in-edges. Then the strongly connected components of
    the directed graph are computed. These strongly connected components are
    unique, regardless of the choice of perfect matching. The strongly connected
    components form a directed acyclic graph, and are returned in a topological
    order. The order is unique, as ambiguities are resolved "lexicographically".

    The "direction" of the projection (where matched edges are out-edges)
    leads to a block *lower* triangular permutation when the top nodes
    correspond to *rows* in the bipartite graph of a matrix.

    Parameters
    ----------
    graph: NetworkX Graph
        A bipartite graph
    top_nodes: list
        One of the bipartite sets in the graph
    matching: dict
        Maps each node in ``top_nodes`` to its matched node

    Returns
    -------
    list of lists
        The outer list is a list of strongly connected components. Each
        strongly connected component is a list of tuples of matched nodes.
        The first node is a "top node", and the second is an "other node".

    """
    nxb = nx.algorithms.bipartite
    nxd = nx.algorithms.dag
    if not nxb.is_bipartite(graph):
        raise RuntimeError("Provided graph is not bipartite.")
    M = len(top_nodes)
    N = len(graph.nodes) - M
    if M != N:
        raise RuntimeError(
            "get_scc_of_projection does not support bipartite graphs with"
            " bipartite sets of different cardinalities. Got sizes %s and"
            " %s." % (M, N)
        )
    if matching is None:
        # This matching maps top nodes to "other nodes" *and* other nodes
        # back to top nodes.
        matching = nxb.maximum_matching(graph, top_nodes=top_nodes)
    if len(matching) != 2 * M:
        raise RuntimeError(
            "get_scc_of_projection does not support bipartite graphs without"
            " a perfect matching. Got a graph with %s nodes per bipartite set"
            " and a matching of cardinality %s." % (M, (len(matching) / 2))
        )

    scc_list, dag = _get_scc_dag_of_projection(graph, top_nodes, matching)
    scc_order = list(nxd.lexicographical_topological_sort(dag))

    # The "natural" return type, here, is a list of lists. Each inner list
    # is an SCC, and contains tuples of nodes. The "top node", and its matched
    # "bottom node".
    ordered_node_subsets = [
        sorted([(i, matching[i]) for i in scc_list[scc_idx]]) for scc_idx in scc_order
    ]
    return ordered_node_subsets


def block_triangularize(matrix, matching=None):
    """Compute ordered partitions of the matrix's rows and columns that
    permute the matrix to block lower triangular form

    Subsets in the partition correspond to diagonal blocks in the block
    triangularization. The order is topological, with ties broken
    "lexicographically".

    Parameters
    ----------
    matrix: ``scipy.sparse.coo_matrix``
        Matrix whose rows and columns will be permuted
    matching: ``dict``
        A perfect matching. Maps rows to columns *and* columns back to rows.

    Returns
    -------
    row_partition: list of lists
        A partition of rows. The inner lists hold integer row coordinates.
    col_partition: list of lists
        A partition of columns. The inner lists hold integer column coordinates.


    .. note::

       **Breaking change in Pyomo 6.5.0**

       The pre-6.5.0 ``block_triangularize`` function returned maps from
       each row or column to the index of its block in a block
       lower triangularization as the original intent of this function
       was to identify when coordinates do or don't share a diagonal block
       in this partition. Since then, the dominant use case of
       ``block_triangularize`` has been to partition variables and
       constraints into these blocks and inspect or solve each block
       individually. A natural return type for this functionality is the
       ordered partition of rows and columns, as lists of lists.
       This functionality was previously available via the
       ``get_diagonal_blocks`` method, which was confusing as it did not
       capture that the partition was the diagonal of a block
       *triangularization* (as opposed to diagonalization). The pre-6.5.0
       functionality of ``block_triangularize`` is still available via the
       ``map_coords_to_block_triangular_indices`` function.

    """
    nxb = nx.algorithms.bipartite
    nxc = nx.algorithms.components
    nxd = nx.algorithms.dag
    from_biadjacency_matrix = nxb.matrix.from_biadjacency_matrix
    M, N = matrix.shape
    if M != N:
        raise ValueError(
            "block_triangularize does not currently support non-square"
            " matrices. Got matrix with shape %s." % ((M, N),)
        )
    graph = from_biadjacency_matrix(matrix)
    row_nodes = list(range(M))
    sccs = get_scc_of_projection(graph, row_nodes, matching=matching)
    row_partition = [[i for i, j in scc] for scc in sccs]
    col_partition = [[j - M for i, j in scc] for scc in sccs]
    return row_partition, col_partition


def map_coords_to_block_triangular_indices(matrix, matching=None):
    row_blocks, col_blocks = block_triangularize(matrix, matching=matching)
    row_idx_map = {r: idx for idx, rblock in enumerate(row_blocks) for r in rblock}
    col_idx_map = {c: idx for idx, cblock in enumerate(col_blocks) for c in cblock}
    return row_idx_map, col_idx_map


@deprecated(
    msg=(
        "``get_blocks_from_maps`` is deprecated. This functionality has been"
        " incorporated into ``block_triangularize``."
    ),
    version="6.5.0",
)
def get_blocks_from_maps(row_block_map, col_block_map):
    blocks = set(row_block_map.values())
    assert blocks == set(col_block_map.values())
    n_blocks = len(blocks)
    block_rows = [[] for _ in range(n_blocks)]
    block_cols = [[] for _ in range(n_blocks)]
    for r, b in row_block_map.items():
        block_rows[b].append(r)
    for c, b in col_block_map.items():
        block_cols[b].append(c)
    return block_rows, block_cols


@deprecated(
    msg=(
        "``get_diagonal_blocks`` has been deprecated. Please use"
        " ``block_triangularize`` instead."
    ),
    version="6.5.0",
)
def get_diagonal_blocks(matrix, matching=None):
    return block_triangularize(matrix, matching=matching)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import argparse
import sys


#
# Sort sub_parser names, since these are inserted throughout Pyomo
#
# NOTE: This may not be robust to different versions of argparse.  We're
# mucking with a non-public API here ...
#
class CustomHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _metavar_formatter(self, action, default_metavar):
        if action.metavar is not None:
            result = action.metavar
        elif action.choices is not None:
            choice_strs = sorted(str(choice) for choice in action.choices)
            result = '{%s}' % ','.join(choice_strs)
        else:
            result = default_metavar

        def format(tuple_size):
            if isinstance(result, tuple):
                return result
            else:
                return (result,) * tuple_size

        return format

    def _iter_indented_subactions(self, action):
        try:
            get_subactions = action._get_subactions
        except AttributeError:
            pass
        else:
            self._indent()
            if isinstance(action, argparse._SubParsersAction):
                for subaction in sorted(get_subactions(), key=lambda x: x.dest):
                    yield subaction
            else:
                for subaction in get_subactions():
                    yield subaction
            self._dedent()


def get_version():
    from pyomo.version import version
    import platform

    return "Pyomo %s (%s %s on %s %s)" % (
        version,
        platform.python_implementation(),
        '.'.join(str(x) for x in sys.version_info[:3]),
        platform.system(),
        platform.release(),
    )


#
# Create the argparse parser for Pyomo
#
doc = "This is the main driver for the Pyomo optimization software."
epilog = """
-------------------------------------------------------------------------
Pyomo supports a variety of modeling and optimization capabilities,
which are executed either as subcommands of 'pyomo' or as separate
commands.  Use the 'help' subcommand to get information about the
capabilities installed with Pyomo.  Additionally, each subcommand
supports independent command-line options.  Use the -h option to
print details for a subcommand.  For example, type

   pyomo solve -h

to print information about the `solve` subcommand.
"""
_pyomo_parser = None
_pyomo_subparsers = None

subparsers = []


def add_subparser(name, **args):
    """
    Add a subparser to the 'pyomo' command.
    """
    if _pyomo_subparsers is None:
        get_parser()

    func = args.pop('func', None)
    parser = _pyomo_subparsers.add_parser(name, **args)
    subparsers.append(name)
    if func is not None:
        parser.set_defaults(func=func)
    return parser


def get_parser():
    """
    Return the parser used by the 'pyomo' command.
    """
    global _pyomo_parser
    if _pyomo_parser is None:
        _pyomo_parser = argparse.ArgumentParser(
            description=doc, epilog=epilog, formatter_class=CustomHelpFormatter
        )
        _pyomo_parser.add_argument("--version", action="version", version=get_version())
        global _pyomo_subparsers
        _pyomo_subparsers = _pyomo_parser.add_subparsers(
            dest='subparser_name', title='subcommands'
        )
    return _pyomo_parser



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.config import document_kwargs_from_configdict
from pyomo.common.errors import DeveloperError
from pyomo.common.modeling import unique_component_name
from pyomo.contrib.gdp_bounds.info import disjunctive_bounds
from pyomo.contrib.gdpopt.algorithm_base_class import _GDPoptAlgorithm
from pyomo.contrib.gdpopt.config_options import (
    _add_oa_configs,
    _add_mip_solver_configs,
    _add_nlp_solver_configs,
    _add_tolerance_configs,
)
from pyomo.contrib.gdpopt.create_oa_subproblems import (
    _get_discrete_problem_and_subproblem,
    add_constraints_by_disjunct,
    add_global_constraint_list,
)
from pyomo.contrib.gdpopt.cut_generation import add_no_good_cut
from pyomo.contrib.gdpopt.oa_algorithm_utils import _OAAlgorithmMixIn
from pyomo.contrib.gdpopt.solve_discrete_problem import solve_MILP_discrete_problem
from pyomo.contrib.gdpopt.util import (
    _add_bigm_constraint_to_transformed_model,
    time_code,
)
from pyomo.contrib.mcpp.pyomo_mcpp import McCormick as mc, MCPP_Error

from pyomo.core import Constraint, Block, NonNegativeIntegers, Objective, value
from pyomo.core.expr.numvalue import is_potentially_variable
from pyomo.core.expr.visitor import identify_variables
from pyomo.opt.base import SolverFactory


@SolverFactory.register(
    'gdpopt.gloa',
    doc="The GLOA (global logic-based outer approximation) Generalized "
    "Disjunctive Programming (GDP) solver",
)
class GDP_GLOA_Solver(_GDPoptAlgorithm, _OAAlgorithmMixIn):
    """The GDPopt (Generalized Disjunctive Programming optimizer) global
    logic-based outer approximation (GLOA) solver.

    Accepts models that can include nonlinear, continuous variables and
    constraints, as well as logical conditions.
    """

    CONFIG = _GDPoptAlgorithm.CONFIG()
    _add_oa_configs(CONFIG)
    _add_mip_solver_configs(CONFIG)
    _add_nlp_solver_configs(CONFIG, default_solver='couenne')
    _add_tolerance_configs(CONFIG)

    algorithm = 'GLOA'

    # Override solve() to customize the docstring for this solver
    @document_kwargs_from_configdict(CONFIG, doc=_GDPoptAlgorithm.solve.__doc__)
    def solve(self, model, **kwds):
        return super().solve(model, **kwds)

    def _log_citation(self, config):
        config.logger.info("\n" + """- GLOA algorithm:
        Lee, S; Grossmann, IE.
        A Global Optimization Algorithm for Nonconvex Generalized
        Disjunctive Programming and Applications to Process Systems.
        Comp. and Chem. Eng. 2001, 25, 1675-1697.
        DOI: 10.1016/S0098-1354(01)00732-3.
        """.strip())

    def _solve_gdp(self, original_model, config):
        logger = config.logger

        # we need to gather a map of Disjuncts to their active Constraints
        # before we call any GDP transformations, as we will need this
        # information for cut generation later
        add_constraints_by_disjunct(self.original_util_block)
        # We also save these in advance because we know only linear logical
        # constraints will be added by the transformation to a MIP, so these are
        # all we'll ever need.
        add_global_constraint_list(self.original_util_block)
        discrete_problem_util_block, subproblem_util_block = (
            _get_discrete_problem_and_subproblem(self, config)
        )
        discrete = discrete_problem_util_block.parent_block()
        subproblem = subproblem_util_block.parent_block()
        discrete_obj = next(
            discrete.component_data_objects(Objective, active=True, descend_into=True)
        )

        self._log_header(logger)

        # main loop
        while not config.iterlim or self.iteration < config.iterlim:
            self.iteration += 1

            # solve linear discrete problem
            with time_code(self.timing, 'mip'):
                mip_feasible = solve_MILP_discrete_problem(
                    discrete_problem_util_block, self, config
                )
                self._update_bounds_after_discrete_problem_solve(
                    mip_feasible, discrete_obj, logger
                )
            # Check termination conditions
            if self.any_termination_criterion_met(config):
                break

            with time_code(self.timing, 'nlp'):
                self._fix_discrete_soln_solve_subproblem_and_add_cuts(
                    discrete_problem_util_block, subproblem_util_block, config
                )

            # Add integer cut
            with time_code(self.timing, "integer cut generation"):
                add_no_good_cut(discrete_problem_util_block, config)

            # Check termination conditions
            if self.any_termination_criterion_met(config):
                break

    def _add_cuts_to_discrete_problem(
        self,
        subproblem_util_block,
        discrete_problem_util_block,
        objective_sense,
        config,
        timing,
    ):
        """Add affine cuts"""
        m = discrete_problem_util_block.parent_block()
        if hasattr(discrete_problem_util_block, "aff_utils_blocks"):
            aff_utils_blocks = discrete_problem_util_block.aff_utils_blocks
        else:
            aff_utils_blocks = discrete_problem_util_block.aff_utils_blocks = dict()

        config.logger.debug("Adding affine cuts.")
        counter = 0
        for discrete_var, subprob_var in zip(
            discrete_problem_util_block.algebraic_variable_list,
            subproblem_util_block.algebraic_variable_list,
        ):
            val = subprob_var.value
            if val is not None and not discrete_var.fixed:
                discrete_var.set_value(val, skip_validation=True)

        for constr in self._get_active_untransformed_constraints(
            discrete_problem_util_block, config
        ):
            disjunctive_var_bounds = disjunctive_bounds(constr.parent_block())

            if constr.body.polynomial_degree() in (1, 0):
                continue

            vars_in_constr = list(identify_variables(constr.body))
            if any(var.value is None for var in vars_in_constr):
                continue  # a variable has no values

            # mcpp stuff
            try:
                mc_eqn = mc(constr.body, disjunctive_var_bounds)
            except MCPP_Error as e:
                config.logger.debug(
                    "Skipping constraint %s due to MCPP "
                    "error %s" % (constr.name, str(e))
                )
                continue  # skip to the next constraint
            ccSlope = mc_eqn.subcc()
            cvSlope = mc_eqn.subcv()
            ccStart = mc_eqn.concave()
            cvStart = mc_eqn.convex()
            ub_int = (
                min(value(constr.upper), mc_eqn.upper())
                if constr.has_ub()
                else mc_eqn.upper()
            )
            lb_int = (
                max(value(constr.lower), mc_eqn.lower())
                if constr.has_lb()
                else mc_eqn.lower()
            )

            parent_block = constr.parent_block()
            # Create a block on which to put outer approximation cuts.
            aff_utils = aff_utils_blocks.get(parent_block)
            if aff_utils is None:
                aff_utils = Block(doc="Block holding affine constraints")
                nm = unique_component_name(parent_block, "GDPopt_aff")
                parent_block.add_component(nm, aff_utils)
                aff_utils_blocks[parent_block] = aff_utils
                aff_utils.GDPopt_aff_cons = Constraint(NonNegativeIntegers)
            aff_cuts = aff_utils.GDPopt_aff_cons
            cut_body = sum(
                ccSlope[var] * (var - var.value)
                for var in vars_in_constr
                if not var.fixed
            )
            if not is_potentially_variable(cut_body):
                if (
                    cut_body + ccStart >= lb_int - config.constraint_tolerance
                    and cut_body + cvStart <= ub_int + config.constraint_tolerance
                ):
                    # We won't add them, but nothing is wrong--they hold
                    config.logger.debug("Affine cut is trivially True.")
                else:
                    # something went wrong.
                    raise DeveloperError("One of the affine cuts is trivially False.")
            else:
                concave_cut = cut_body + ccStart >= lb_int
                convex_cut = cut_body + cvStart <= ub_int
                idx = len(aff_cuts)
                aff_cuts[idx] = concave_cut
                aff_cuts[idx + 1] = convex_cut
                _add_bigm_constraint_to_transformed_model(m, aff_cuts[idx], aff_cuts)
                _add_bigm_constraint_to_transformed_model(
                    m, aff_cuts[idx + 1], aff_cuts
                )
                counter += 2

                config.logger.debug("Added %s affine cuts" % counter)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

logger = logging.getLogger('pyomo.core')

from pyomo.common import deprecated
from pyomo.common.config import ConfigDict, ConfigValue, In, IsInstance
from pyomo.common.deprecation import deprecation_warning
from pyomo.core.base import (
    Transformation,
    TransformationFactory,
    Var,
    Suffix,
    Reals,
    Block,
    ReverseTransformationToken,
    VarCollector,
    Constraint,
    Objective,
)
from pyomo.core.base.block import SubclassOf
from pyomo.core.util import target_list
from pyomo.gdp import Disjunct
from pyomo.util.vars_from_expressions import get_vars_from_components


#
# This transformation relaxes integer ranges to their continuous
# counterparts
#
@TransformationFactory.register(
    'core.relax_integer_vars', doc="Relax integer variables to continuous counterparts"
)
class RelaxIntegerVars(Transformation):
    CONFIG = ConfigDict('core.relax_integer_vars')
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets that will be relaxed",
            doc="""
            This specifies the list of components to relax. If None (default), the
            entire model is transformed. Note that if the transformation is done
            out of place, the list of targets should be attached to the model before
            it is cloned, and the list will specify the targets on the cloned
            instance.""",
        ),
    )
    CONFIG.declare(
        'reverse',
        ConfigValue(
            default=None,
            domain=IsInstance(ReverseTransformationToken),
            description="The token returned by a (forward) call to this "
            "transformation, if you wish to reverse the transformation.",
            doc="""
            This argument should be the reverse transformation token
            returned by a previous call to this transformation to transform
            fixed discrete state in the given model.
            If this argument is specified, this call to the transformation
            will reverse what the transformation did in the call that returned
            the token. Note that if there are intermediate changes to the model
            in between the forward and the backward calls to the transformation,
            the behavior could be unexpected.
            """,
        ),
    )
    CONFIG.declare(
        'var_collector',
        ConfigValue(
            default=VarCollector.FromVarComponents,
            domain=In(VarCollector),
            description="The method for collection the Vars to relax. If "
            "VarCollector.FromVarComponents (default), any Var component on "
            "the active tree will be relaxed.",
            doc="""
            This specifies the method for collecting the Var components to relax.
            The default, VarCollector.FromVarComponents, assumes that all relevant
            Vars are on the active tree. If this is true, then this is the most
            performant option. However, in more complex cases where some Vars may not
            be in the active tree (e.g. some are on deactivated Blocks or come from
            other models), specify VarCollector.FromExpressions to relax all Vars that
            appear in expressions in the active tree.
            """,
        ),
    )
    CONFIG.declare(
        'transform_deactivated_blocks',
        ConfigValue(
            default=True,
            description="[DEPRECATED]: Whether or not to search for Var components to "
            "relax on deactivated Blocks. True by default",
        ),
    )
    CONFIG.declare(
        'undo',
        ConfigValue(
            default=False,
            domain=bool,
            description="[DEPRECATED]: Please use the 'reverse' argument to undo "
            "the transformation.",
        ),
    )

    def __init__(self):
        super().__init__()

    def _apply_to(self, model, **kwds):
        if model.ctype not in SubclassOf(Block):
            raise ValueError(
                "Transformation called on %s of type %s. 'model' "
                "must be a ConcreteModel or Block." % (model.name, model.ctype)
            )
        config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)

        if config.undo:
            deprecation_warning(
                "The 'undo' argument is deprecated. Please use the 'reverse' "
                "argument to undo the transformation.",
                version='6.9.3',
            )
            for v, d in model._relaxed_integer_vars[None].values():
                bounds = v.bounds
                v.domain = d
                v.setlb(bounds[0])
                v.setub(bounds[1])
            model.del_component("_relaxed_integer_vars")
            return

        targets = (model,) if config.targets is None else config.targets

        if config.reverse is None:
            reverse_dict = {}
            # Relax the model
            reverse_token = ReverseTransformationToken(
                self.__class__, model, targets, reverse_dict
            )
        else:
            # reverse the transformation
            reverse_token = config.reverse
            reverse_token.check_token_valid(self.__class__, model, targets)
            reverse_dict = reverse_token.reverse_dict
            for v, d in reverse_dict.values():
                lb, ub = v.bounds
                v.domain = d
                v.setlb(lb)
                v.setub(ub)
            return

        ### [ESJ 4/29/25]: This can go away when we remove 'undo'
        model._relaxed_integer_vars = Suffix(direction=Suffix.LOCAL)
        model._relaxed_integer_vars[None] = reverse_dict
        ###

        for t in targets:
            if isinstance(t, Block):
                blocks = t.values() if t.is_indexed() else (t,)
                for block in blocks:
                    self._relax_block(block, config, reverse_dict)
            elif t.ctype is Var:
                self._relax_var(t, reverse_dict)
            else:
                raise ValueError(
                    "Target '%s' was not a Block or Var. It was of type "
                    "'%s' and cannot be transformed." % (t.name, type(t))
                )

        return reverse_token

    def _relax_block(self, block, config, reverse_dict):
        self._relax_vars_from_block(block, config, reverse_dict)

        for b in block.component_data_objects(Block, active=None, descend_into=True):
            if not b.active:
                if config.transform_deactivated_blocks:
                    deprecation_warning(
                        "The `transform_deactivated_blocks` arguments is deprecated. "
                        "Either specify deactivated Blocks as targets to activate them "
                        "if transforming them is the desired behavior.",
                        version='6.9.3',
                    )
                else:
                    continue
            self._relax_vars_from_block(b, config, reverse_dict)

    def _relax_vars_from_block(self, block, config, reverse_dict):
        if config.var_collector is VarCollector.FromVarComponents:
            model_vars = block.component_data_objects(Var, descend_into=False)
        else:
            model_vars = get_vars_from_components(
                block, ctype=(Constraint, Objective), descend_into=False
            )
        for var in model_vars:
            if id(var) not in reverse_dict:
                self._relax_var(var, reverse_dict)

    def _relax_var(self, v, reverse_dict):
        var_datas = v.values() if v.is_indexed() else (v,)
        for var in var_datas:
            if not var.is_integer():
                continue
            lb, ub = var.bounds
            _domain = var.domain
            var.domain = Reals
            var.setlb(lb)
            var.setub(ub)
            reverse_dict[id(var)] = (var, _domain)


@TransformationFactory.register(
    'core.relax_discrete',
    doc="[DEPRECATED] Relax integer variables to continuous counterparts",
)
@deprecated(
    "core.relax_discrete is deprecated.  Use core.relax_integer_vars", version='5.7'
)
class RelaxDiscreteVars(RelaxIntegerVars):
    """
    This plugin relaxes integrality in a Pyomo model.
    """

    def __init__(self, **kwds):
        super(RelaxDiscreteVars, self).__init__(**kwds)


#
# This transformation fixes known discrete domains to their current values
#
@TransformationFactory.register(
    'core.fix_integer_vars', doc="Fix all integer variables to their current values"
)
class FixIntegerVars(Transformation):
    def __init__(self):
        super(FixIntegerVars, self).__init__()

    def _apply_to(self, model, **kwds):
        options = kwds.pop('options', {})
        if kwds.get('undo', options.get('undo', False)):
            for v in model._fixed_integer_vars[None]:
                v.unfix()
            model.del_component("_fixed_integer_vars")
            return

        fixed_vars = []
        _base_model_vars = model.component_data_objects(
            Var, active=True, descend_into=True
        )
        for var in _base_model_vars:
            # Instead of checking against
            # `_integer_relaxation_map.keys()` we just check the item
            # properties to fix #995 When #326 has been resolved, we can
            # check against the dict-keys again
            if var.is_integer() and not var.is_fixed():
                fixed_vars.append(var)
                var.fix()
        model._fixed_integer_vars = Suffix(direction=Suffix.LOCAL)
        model._fixed_integer_vars[None] = fixed_vars


@TransformationFactory.register(
    'core.fix_discrete',
    doc="[DEPRECATED] Fix all integer variables to their current values",
)
@deprecated(
    "core.fix_discrete is deprecated.  Use core.fix_integer_vars", version='5.7'
)
class FixDiscreteVars(FixIntegerVars):
    def __init__(self, **kwds):
        super(FixDiscreteVars, self).__init__(**kwds)



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Transformation to deactivate trivial constraints."""

import logging

from pyomo.common.collections import ComponentSet
from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    NonNegativeFloat,
    document_kwargs_from_configdict,
)
from pyomo.common.errors import InfeasibleConstraintException
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.transformation import TransformationFactory
from pyomo.core.expr.numvalue import value
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.repn import generate_standard_repn

logger = logging.getLogger('pyomo.contrib.preprocessing')


@TransformationFactory.register(
    'contrib.deactivate_trivial_constraints', doc="Deactivate trivial constraints."
)
@document_kwargs_from_configdict('CONFIG')
class TrivialConstraintDeactivator(IsomorphicTransformation):
    """Deactivates trivial constraints.

    Trivial constraints take form :math:`k_1 = k_2` or :math:`k_1 \\leq k_2`,
    where :math:`k_1` and :math:`k_2` are constants. These constraints
    typically arise when variables are fixed.

    Keyword arguments below are specified for the ``apply_to`` and
    ``create_using`` functions.

    """

    CONFIG = ConfigBlock("TrivialConstraintDeactivator")
    CONFIG.declare(
        "tmp",
        ConfigValue(
            default=False,
            domain=bool,
            description="True to store a set of transformed constraints for future"
            " reversion of the transformation.",
        ),
    )
    CONFIG.declare(
        "ignore_infeasible",
        ConfigValue(
            default=False,
            domain=bool,
            description="True to skip over trivial constraints that are "
            "infeasible instead of raising an InfeasibleConstraintException.",
        ),
    )
    CONFIG.declare(
        "return_trivial",
        ConfigValue(
            default=[],
            description="a list to which the deactivated trivial"
            "constraints are appended (side effect)",
        ),
    )
    CONFIG.declare(
        "tolerance",
        ConfigValue(
            default=1e-13,
            domain=NonNegativeFloat,
            description="tolerance on constraint violations",
        ),
    )

    def _apply_to(self, instance, **kwargs):
        config = self.CONFIG(kwargs)
        if config.tmp and not hasattr(instance, '_tmp_trivial_deactivated_constrs'):
            instance._tmp_trivial_deactivated_constrs = ComponentSet()
        elif config.tmp:
            logger.warning(
                'Deactivating trivial constraints on the block {} for which '
                'trivial constraints were previously deactivated. '
                'Reversion will affect all deactivated constraints.'.format(
                    instance.name
                )
            )

        # Trivial constraints are those that do not contain any variables, ie.
        # the polynomial degree is 0
        for constr in instance.component_data_objects(
            ctype=Constraint, active=True, descend_into=True
        ):
            repn = generate_standard_repn(constr.body)
            if not repn.is_constant():
                # This constraint is not trivial
                continue

            # We need to check each constraint to sure that it is not violated.
            constr_lb = value(constr.lower) if constr.has_lb() else float('-inf')
            constr_ub = value(constr.upper) if constr.has_ub() else float('inf')
            constr_value = repn.constant

            # Check if the lower bound is violated outside a given tolerance
            if constr_value + config.tolerance <= constr_lb:
                if config.ignore_infeasible:
                    continue
                else:
                    raise InfeasibleConstraintException(
                        'Trivial constraint {} violates LB {} ≤ BODY {}.'.format(
                            constr.name, constr_lb, constr_value
                        )
                    )

            # Check if the upper bound is violated outside a given tolerance
            if constr_value >= constr_ub + config.tolerance:
                if config.ignore_infeasible:
                    continue
                else:
                    raise InfeasibleConstraintException(
                        'Trivial constraint {} violates BODY {} ≤ UB {}.'.format(
                            constr.name, constr_value, constr_ub
                        )
                    )

            # Constraint is not infeasible. Deactivate it.
            if config.tmp:
                instance._tmp_trivial_deactivated_constrs.add(constr)
            config.return_trivial.append(constr)
            constr.deactivate()

    def revert(self, instance):
        """Revert constraints deactivated by the transformation.

        Args:
            instance: the model instance on which trivial constraints were
                earlier deactivated.
        """
        for constr in instance._tmp_trivial_deactivated_constrs:
            constr.activate()
        del instance._tmp_trivial_deactivated_constrs



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.environ import (
    Param,
    Var,
    Block,
    ComponentMap,
    Objective,
    Constraint,
    ConstraintList,
    Suffix,
    value,
    ComponentUID,
)

from pyomo.common.sorting import sorted_robust
from pyomo.core.expr import ExpressionReplacementVisitor
from pyomo.core.expr.numvalue import is_potentially_variable

from pyomo.common.modeling import unique_component_name
from pyomo.common.dependencies import numpy as np, scipy
from pyomo.common.deprecation import deprecated
from pyomo.common.tempfiles import TempfileManager
from pyomo.opt import SolverFactory, SolverStatus
from pyomo.contrib.sensitivity_toolbox.k_aug import K_augInterface, InTempDir
import logging
import os
import io
import shutil

logger = logging.getLogger('pyomo.contrib.sensitivity_toolbox')


@deprecated(
    "The sipopt function has been deprecated. Use the sensitivity_calculation() "
    "function with method='sipopt' to access this functionality.",
    logger='pyomo.contrib.sensitivity_toolbox',
    version='6.1',
)
def sipopt(
    instance,
    paramSubList,
    perturbList,
    cloneModel=True,
    tee=False,
    keepfiles=False,
    streamSoln=False,
):
    m = sensitivity_calculation(
        'sipopt',
        instance,
        paramSubList,
        perturbList,
        cloneModel,
        tee,
        keepfiles,
        solver_options=None,
    )

    return m


@deprecated(
    "The kaug function has been deprecated. Use the sensitivity_calculation() "
    "function with method='k_aug' to access this functionality.",
    logger='pyomo.contrib.sensitivity_toolbox',
    version='6.1',
)
def kaug(
    instance,
    paramSubList,
    perturbList,
    cloneModel=True,
    tee=False,
    keepfiles=False,
    solver_options=None,
    streamSoln=False,
):
    m = sensitivity_calculation(
        'k_aug',
        instance,
        paramSubList,
        perturbList,
        cloneModel,
        tee,
        keepfiles,
        solver_options,
    )

    return m


_SIPOPT_SUFFIXES = {
    'sens_state_0': Suffix.EXPORT,
    # ^ Not sure what this suffix does -RBP
    'sens_state_1': Suffix.EXPORT,
    'sens_state_value_1': Suffix.EXPORT,
    'sens_init_constr': Suffix.EXPORT,
    'sens_sol_state_1': Suffix.IMPORT,
    'sens_sol_state_1_z_L': Suffix.IMPORT,
    'sens_sol_state_1_z_U': Suffix.IMPORT,
}

_K_AUG_SUFFIXES = {
    'ipopt_zL_out': Suffix.IMPORT,
    'ipopt_zU_out': Suffix.IMPORT,
    'ipopt_zL_in': Suffix.EXPORT,
    'ipopt_zU_in': Suffix.EXPORT,
    'dual': Suffix.IMPORT_EXPORT,
    'dcdp': Suffix.EXPORT,
    'DeltaP': Suffix.EXPORT,
}


def _add_sensitivity_suffixes(block):
    suffix_dict = {}
    suffix_dict.update(_SIPOPT_SUFFIXES)
    suffix_dict.update(_K_AUG_SUFFIXES)
    for name, direction in suffix_dict.items():
        if block.component(name) is None:
            # Only add suffix if it doesn't already exist.
            # If something of this name does already exist, just
            # assume it is the proper suffix and move on.
            block.add_component(name, Suffix(direction=direction))


class _NotAnIndex:
    pass


def _generate_component_items(components):
    if type(components) not in {list, tuple}:
        components = (components,)
    for comp in components:
        if comp.is_indexed():
            for idx in sorted_robust(comp):
                yield idx, comp[idx]
        else:
            yield _NotAnIndex, comp


def sensitivity_calculation(
    method,
    instance,
    paramList,
    perturbList,
    cloneModel=True,
    tee=False,
    keepfiles=False,
    solver_options=None,
):
    """This function accepts a Pyomo ConcreteModel, a list of parameters, and
    their corresponding perturbation list. The model is then augmented with
    dummy constraints required to call sipopt or k_aug to get an approximation
    of the perturbed solution.

    Parameters
    ----------
    method: string
        'sipopt' or 'k_aug'
    instance: Block
        pyomo block or model object
    paramSubList: list
        list of mutable parameters or fixed variables
    perturbList: list
        list of perturbed parameter values
    cloneModel: bool, optional
        indicator to clone the model. If set to False, the original
        model will be altered
    tee: bool, optional
        indicator to stream solver log
    keepfiles: bool, optional
        preserve solver interface files
    solver_options: dict, optional
        Provides options to the solver (also the name of an attribute)

    Returns
    -------
    The model that was manipulated by the sensitivity interface

    """

    sens = SensitivityInterface(instance, clone_model=cloneModel)
    sens.setup_sensitivity(paramList)

    m = sens.model_instance

    if method not in {"k_aug", "sipopt"}:
        raise ValueError("Only methods 'k_aug' and 'sipopt' are supported'")

    if method == 'k_aug':
        k_aug = SolverFactory('k_aug', solver_io='nl')
        dot_sens = SolverFactory('dot_sens', solver_io='nl')
        ipopt = SolverFactory('ipopt', solver_io='nl')

        k_aug_interface = K_augInterface(k_aug=k_aug, dot_sens=dot_sens)

        ipopt.solve(m, tee=tee)
        m.ipopt_zL_in.update(m.ipopt_zL_out)  #: important!
        m.ipopt_zU_in.update(m.ipopt_zU_out)  #: important!

        k_aug.options['dsdp_mode'] = ""  #: sensitivity mode!
        k_aug_interface.k_aug(m, tee=tee)

    sens.perturb_parameters(perturbList)

    if method == 'sipopt':
        # Notes on sIpopt documentation:
        # Documentation:
        #  - https://coin-or.github.io/Ipopt/SPECIALS.html#SIPOPT
        # Original docs (archived):
        #  - http://web.archive.org/web/20210412132144/https://projects.coin-or.org/Ipopt/wiki/sIpopt
        ipopt_sens = SolverFactory('ipopt_sens', solver_io='nl')
        ipopt_sens.options['run_sens'] = 'yes'
        if solver_options is not None:
            ipopt_sens.options['linear_solver'] = solver_options

        # Send the model to ipopt_sens and collect the solution
        results = ipopt_sens.solve(m, keepfiles=keepfiles, tee=tee)

    elif method == 'k_aug':
        dot_sens.options["dsdp_mode"] = ""
        k_aug_interface.dot_sens(m, tee=tee)

    return m


def get_dsdp(model, theta_names, theta, tee=False):
    r"""This function calculates gradient vector of the variables with
    respect to the parameters (theta_names).

    For example, given:

    .. math::

        \min f:\ & p1*x1 + p2*(x2^2) + p1*p2 \\
        s.t.\ & c1: x1 + x2 = p1 \\
              & c2: x2 + x3 = p2 \\
              & 0 <= x1, x2, x3 <= 10 \\
              & p1 = 10 \\
              & p2 = 5

    the function returns dx/dp and dp/dp, and column orders.

    The following terms are used to define the output dimensions:
    - Ncon   = number of constraints
    - Nvar   = number of variables (Nx + Ntheta)
    - Nx     = number of decision (primal) variables
    - Ntheta = number of uncertain parameters.

    Parameters
    ----------
    model: Pyomo ConcreteModel
        model should include an objective function
    theta_names: list of strings
        List of Var names
    theta: dict
        Estimated parameters e.g) from parmest
    tee: bool, optional
        Indicates that ef solver output should be teed

    Returns
    -------
    dsdp: scipy.sparse.csr.csr_matrix
        Ntheta by Nvar size sparse matrix. A Jacobian matrix of the
        (decision variables, parameters) with respect to parameters
        (theta_names). number of rows = len(theta_name), number of
        columns = len(col)
    col: list
        List of variable names

    """
    # Get parameters from names. In SensitivityInterface, we expect
    # these to be parameters on the original model.
    param_list = []
    for name in theta_names:
        comp = model.find_component(name)
        if comp is None:
            raise RuntimeError("Cannot find component %s on model" % name)
        if comp.ctype is Var:
            # If theta_names correspond to Vars in the model, these vars
            # need to be fixed.
            comp.fix()
        param_list.append(comp)

    sens = SensitivityInterface(model, clone_model=True)
    m = sens.model_instance

    # Setup model and calculate sensitivity matrix with k_aug
    sens.setup_sensitivity(param_list)
    k_aug = K_augInterface()
    k_aug.k_aug(m, tee=tee)

    # Write row and col files in a temp dir, then immediately
    # read into a Python data structure.
    nl_data = {}
    with InTempDir():
        base_fname = "col_row"
        nl_file = ".".join((base_fname, "nl"))
        row_file = ".".join((base_fname, "row"))
        col_file = ".".join((base_fname, "col"))
        m.write(nl_file, io_options={"symbolic_solver_labels": True})
        for fname in [nl_file, row_file, col_file]:
            with open(fname, "r") as fp:
                nl_data[fname] = fp.read()

    # Create more useful data structures from strings
    dsdp = np.fromstring(k_aug.data["dsdp_in_.in"], sep="\n\t")
    col = nl_data[col_file].strip("\n").split("\n")
    row = nl_data[row_file].strip("\n").split("\n")

    dsdp = dsdp.reshape((len(theta_names), int(len(dsdp) / len(theta_names))))
    dsdp = dsdp[: len(theta_names), : len(col)]

    col = [i for i in col if sens.get_default_block_name() not in i]
    dsdp_out = np.zeros((len(theta_names), len(col)))
    for i in range(len(theta_names)):
        for j in range(len(col)):
            if sens.get_default_block_name() not in col[j]:
                dsdp_out[i, j] = -dsdp[i, j]  # e.g) k_aug dsdp returns -dx1/dx1 = -1.0

    return scipy.sparse.csr_matrix(dsdp_out), col


def get_dfds_dcds(model, theta_names, tee=False, solver_options=None):
    r"""This function calculates gradient vector of the objective function
    and constraints with respect to the variables and parameters.

    For example, given:

    .. math::

        \min f:\ & p1*x1 + p2*(x2^2) + p1*p2 \\
        s.t.\ & c1: x1 + x2 = p1 \\
              & c2: x2 + x3 = p2 \\
              & 0 <= x1, x2, x3 <= 10 \\
              & p1 = 10 \\
              & p2 = 5

    - Variables = (x1, x2, x3, p1, p2)
    - Fix p1 and p2 with estimated values

    The following terms are used to define the output dimensions:
    - Ncon   = number of constraints
    - Nvar   = number of variables (Nx + Ntheta)
    - Nx     = number of decision (primal) variables
    - Ntheta = number of uncertain parameters.

    Parameters
    ----------
    model : Pyomo ConcreteModel
        model should include an objective function

    theta_names : list of strings
        List of Var names

    tee : bool, optional
        Indicates that ef solver output should be teed

    solver_options : dict, optional
        Provides options to the solver (also the name of an attribute)

    Returns
    -------
    gradient_f : numpy.ndarray
        Length Nvar array. A gradient vector of the objective function
        with respect to the (decision variables, parameters) at the optimal
        solution

    gradient_c : scipy.sparse.csr.csr_matrix
        Ncon by Nvar size sparse matrix. A Jacobian matrix of the
        constraints with respect to the (decision variables, parameters)
        at the optimal solution. Each row contains [row number,
        column number, and value], column order follows variable order in col
        and index starts from 0. Note that it follows k_aug.
        If no constraint exists, return []

    col : list
        Size Nvar list of variable names

    row : list
        Size Ncon+1 list of constraints and objective function names.
        The final element is the objective function name.

    line_dic : dict
        column numbers of the theta_names in the model. Index starts from 1

    Raises
    ------
    RuntimeError
        When ipopt or k_aug or dotsens is not available
    Exception
        When ipopt fails
    """
    # Create the solver plugin using the ASL interface
    ipopt = SolverFactory('ipopt', solver_io='nl')
    if solver_options is not None:
        ipopt.options = solver_options
    k_aug = SolverFactory('k_aug', solver_io='nl')
    if not ipopt.available(False):
        raise RuntimeError('ipopt is not available')
    if not k_aug.available(False):
        raise RuntimeError('k_aug is not available')

    # Declare Suffixes
    _add_sensitivity_suffixes(model)

    # K_AUG SUFFIXES
    model.dof_v = Suffix(direction=Suffix.EXPORT)  #: SUFFIX FOR K_AUG
    model.rh_name = Suffix(direction=Suffix.IMPORT)  #: SUFFIX FOR K_AUG AS WELL
    k_aug.options["print_kkt"] = ""

    results = ipopt.solve(model, tee=tee)

    # Raise exception if ipopt fails
    if results.solver.status == SolverStatus.warning:
        raise Exception(results.solver.Message)

    for o in model.component_objects(Objective, active=True):
        f_mean = value(o)
    model.ipopt_zL_in.update(model.ipopt_zL_out)
    model.ipopt_zU_in.update(model.ipopt_zU_out)

    # run k_aug
    k_aug_interface = K_augInterface(k_aug=k_aug)
    k_aug_interface.k_aug(model, tee=tee)  #: always call k_aug AFTER ipopt.

    nl_data = {}
    with InTempDir():
        base_fname = "col_row"
        nl_file = ".".join((base_fname, "nl"))
        row_file = ".".join((base_fname, "row"))
        col_file = ".".join((base_fname, "col"))
        model.write(nl_file, io_options={"symbolic_solver_labels": True})
        for fname in [nl_file, row_file, col_file]:
            with open(fname, "r") as fp:
                nl_data[fname] = fp.read()

    col = nl_data[col_file].strip("\n").split("\n")
    row = nl_data[row_file].strip("\n").split("\n")

    # get the column numbers of "parameters"
    line_dic = {name: col.index(name) for name in theta_names}

    grad_f_file = os.path.join("GJH", "gradient_f_print.txt")
    grad_f_string = k_aug_interface.data[grad_f_file]
    gradient_f = np.fromstring(grad_f_string, sep="\n\t")
    col = [i for i in col if SensitivityInterface.get_default_block_name() not in i]

    grad_c_file = os.path.join("GJH", "A_print.txt")
    grad_c_string = k_aug_interface.data[grad_c_file]
    gradient_c = np.fromstring(grad_c_string, sep="\n\t")

    # Jacobian file is in "COO format," i.e. an nnz-by-3 array.
    # Reshape to a numpy array that matches this format.
    gradient_c = gradient_c.reshape((-1, 3))

    num_constraints = len(row) - 1  # Objective is included as a row
    if num_constraints > 0:
        row_idx = gradient_c[:, 1] - 1
        col_idx = gradient_c[:, 0] - 1
        data = gradient_c[:, 2]
        gradient_c = scipy.sparse.csr_matrix(
            (data, (row_idx, col_idx)), shape=(num_constraints, len(col))
        )
    else:
        gradient_c = np.array([])

    return gradient_f, gradient_c, col, row, line_dic


def line_num(file_name, target):
    """This function returns the line number that contains 'target' in the
    file_name. This function identifies constraints that have variables
    in theta_names.

    Parameters
    ----------
    file_name: string
        file includes the variable order (i.e. col file)
    target: string
        variable name to check

    Returns
    -------
    count: int
        line number of target in the file

    Raises
    ------
    Exception
        When file does not include target
    """
    with open(file_name) as f:
        count = int(1)
        for line in f:
            if line.strip() == target:
                return int(count)
            count += 1
    raise Exception(file_name + " does not include " + target)


class SensitivityInterface:
    def __init__(self, instance, clone_model=True):
        """Constructor clones model if necessary and attaches
        to this object.
        """
        self._original_model = instance

        if clone_model:
            # Note that we are not "cloning" the user's parameters
            # or perturbations.
            self.model_instance = instance.clone()
        else:
            self.model_instance = instance

    @classmethod
    def get_default_block_name(self):
        return '_SENSITIVITY_TOOLBOX_DATA'

    @staticmethod
    def get_default_var_name(name):
        # return '_'.join(('sens_var', name))
        return name

    @staticmethod
    def get_default_param_name(name):
        # return '_'.join(('sens_param', name))
        return name

    def _process_param_list(self, paramList):
        # We need to translate the components in paramList into
        # components in our possibly cloned model.
        orig = self._original_model
        instance = self.model_instance
        if orig is not instance:
            paramList = list(
                ComponentUID(param, context=orig).find_component_on(instance)
                for param in paramList
            )
        return paramList

    def _add_data_block(self, existing_block=None):
        # If a sensitivity block already exists, and we have not done
        # any expression replacement, we delete the old block, re-fix the
        # sensitivity variables, and start again.
        #
        # Don't do this in the constructor as we could want to call
        # the constructor once, then perform multiple sensitivity
        # calculations with the same model instance.
        if existing_block is not None:
            if (
                hasattr(existing_block, '_has_replaced_expressions')
                and not existing_block._has_replaced_expressions
            ):
                for var, _, _, _ in existing_block._sens_data_list:
                    # Re-fix variables that the previous block was
                    # treating as parameters.
                    var.fix()
                self.model_instance.del_component(existing_block)
            else:
                msg = (
                    "Re-using sensitivity interface is not supported "
                    "when calculating sensitivity for mutable parameters. "
                    "Used fixed vars instead if you want to do this."
                )
                raise RuntimeError(msg)

        # Add a block to keep track of model components necessary for this
        # sensitivity calculation.
        block = Block()
        self.model_instance.add_component(self.get_default_block_name(), block)
        self.block = block

        # If the user tells us they will perturb a set of parameters, we will
        # need to replace these parameters in the user's model's constraints.
        # This affects what we can do with the model later, so we add a flag.
        block._has_replaced_expressions = False

        # This is the main data structure for keeping track of "sensitivity
        # vars" and their associated params. It will be a list of tuples of
        # (vardata, paramdata, list_index, comp_index)
        # where:
        #     vardata is the "sensitivity variable" data object,
        #     paramdata is the associated parameter,
        #     list_index is its index in the user-provided list, and
        #     comp_index is its index in the component provided by the user.
        block._sens_data_list = []

        # This will hold the user-provided list of
        # variables and/or parameters to perturb
        block._paramList = None

        # This will hold any constraints where we have replaced
        # parameters with variables.
        block.constList = ConstraintList()

        return block

    def _add_sensitivity_data(self, param_list):
        block = self.block
        sens_data_list = block._sens_data_list
        for i, comp in enumerate(param_list):
            if comp.ctype is Param:
                parent = comp.parent_component()
                if not parent.mutable:
                    raise ValueError(
                        "Parameters within paramList must be mutable. "
                        "Got %s, which is not mutable." % comp.name
                    )
                # Add a Var:
                if comp.is_indexed():
                    d = {k: value(comp[k]) for k in comp.index_set()}
                    var = Var(comp.index_set(), initialize=d)
                else:
                    d = value(comp)
                    var = Var(initialize=d)
                name = self.get_default_var_name(parent.local_name)
                name = unique_component_name(block, name)
                block.add_component(name, var)

                if comp.is_indexed():
                    sens_data_list.extend(
                        (var[idx], param, i, idx)
                        for idx, param in _generate_component_items(comp)
                    )
                else:
                    sens_data_list.append((var, comp, i, _NotAnIndex))

            elif comp.ctype is Var:
                parent = comp.parent_component()
                for _, data in _generate_component_items(comp):
                    if not data.fixed:
                        raise ValueError(
                            "Specified \"parameter\" variables must be "
                            "fixed. Got %s, which is not fixed." % comp.name
                        )
                # Add a Param:
                if comp.is_indexed():
                    d = {k: value(comp[k]) for k in comp.index_set()}
                    param = Param(comp.index_set(), mutable=True, initialize=d)
                else:
                    d = value(comp)
                    param = Param(mutable=True, initialize=d)
                name = self.get_default_param_name(parent.local_name)
                name = unique_component_name(block, name)
                block.add_component(name, param)

                if comp.is_indexed():
                    sens_data_list.extend(
                        (var, param[idx], i, idx)
                        for idx, var in _generate_component_items(comp)
                    )
                else:
                    sens_data_list.append((comp, param, i, _NotAnIndex))

    def _replace_parameters_in_constraints(self, variableSubMap):
        instance = self.model_instance
        block = self.block
        # Visitor that we will use to replace user-provided parameters
        # in the objective and the constraints.
        param_replacer = ExpressionReplacementVisitor(
            substitute=variableSubMap, remove_named_expressions=True
        )
        # TODO: Flag to ExpressionReplacementVisitor to only replace
        # named expressions if a node has been replaced within that
        # expression.

        new_old_comp_map = ComponentMap()

        # clone Objective, add to Block, and update any Expressions
        for obj in list(
            instance.component_data_objects(Objective, active=True, descend_into=True)
        ):
            tempName = unique_component_name(block, obj.local_name)
            new_expr = param_replacer.walk_expression(obj.expr)
            block.add_component(tempName, Objective(expr=new_expr))
            new_old_comp_map[block.component(tempName)] = obj
            obj.deactivate()

        # clone Constraints, add to Block, and update any Expressions
        #
        # Unfortunate that this deactivates and replaces constraints
        # even if they don't contain the parameters.
        #
        old_con_list = list(
            instance.component_data_objects(Constraint, active=True, descend_into=True)
        )
        last_idx = 0
        for con in old_con_list:
            new_expr = param_replacer.walk_expression(con.expr)
            # TODO: We could only create new constraints for expressions
            # where substitution actually happened, but that breaks some
            # current tests:
            #
            # if new_expr is con.expr:
            #     # No params were substituted.  We can ignore this constraint
            #     continue
            if new_expr.nargs() == 3 and (
                is_potentially_variable(new_expr.arg(0))
                or is_potentially_variable(new_expr.arg(2))
            ):
                # This is a potentially "invalid" range constraint: it
                # may now have variables in the bounds.  For safety, we
                # will split it into two simple inequalities.
                block.constList.add(expr=(new_expr.arg(0) <= new_expr.arg(1)))
                last_idx += 1
                new_old_comp_map[block.constList[last_idx]] = con
                block.constList.add(expr=(new_expr.arg(1) <= new_expr.arg(2)))
                last_idx += 1
                new_old_comp_map[block.constList[last_idx]] = con
            else:
                block.constList.add(expr=new_expr)
                last_idx += 1
                new_old_comp_map[block.constList[last_idx]] = con
            con.deactivate()

        return new_old_comp_map

    def setup_sensitivity(self, paramList):
        instance = self.model_instance
        paramList = self._process_param_list(paramList)

        existing_block = instance.component(self.get_default_block_name())
        block = self._add_data_block(existing_block=existing_block)
        block._sens_data_list = []
        block._paramList = paramList

        self._add_sensitivity_data(paramList)
        sens_data_list = block._sens_data_list

        for var, _, _, _ in sens_data_list:
            # This unfixes all variables, not just those the user added.
            var.unfix()

        # Map used to replace user-provided parameters.
        variableSubMap = dict(
            (id(param), var)
            for var, param, list_idx, _ in sens_data_list
            if paramList[list_idx].ctype is Param
        )

        if variableSubMap:
            # We now replace the provided parameters in the user's
            # expressions. Only do this if we have to, i.e. the
            # user provided some parameters rather than all vars.
            block._replaced_map = self._replace_parameters_in_constraints(
                variableSubMap
            )

            # Assume that we just replaced some params
            block._has_replaced_expressions = True

        block.paramConst = ConstraintList()
        for var, param, _, _ in sens_data_list:
            # block.paramConst.add(param - var == 0)
            block.paramConst.add(var - param == 0)

        # Declare Suffixes
        _add_sensitivity_suffixes(instance)

        for i, (var, _, _, _) in enumerate(sens_data_list):
            idx = i + 1
            con = block.paramConst[idx]

            # sipopt
            instance.sens_state_0[var] = idx
            instance.sens_state_1[var] = idx
            instance.sens_init_constr[con] = idx

            # k_aug
            instance.dcdp[con] = idx

    def perturb_parameters(self, perturbList):
        # Note that entries of perturbList need not be components
        # of the cloned model. All we need are the values.
        instance = self.model_instance
        sens_data_list = self.block._sens_data_list
        paramConst = self.block.paramConst

        if len(self.block._paramList) != len(perturbList):
            raise ValueError(
                "Length of paramList argument does not equal length of perturbList"
            )

        for i, (var, param, list_idx, comp_idx) in enumerate(sens_data_list):
            con = paramConst[i + 1]
            if comp_idx is _NotAnIndex:
                ptb = value(perturbList[list_idx])
            else:
                try:
                    ptb = value(perturbList[list_idx][comp_idx])
                except TypeError:
                    # If the user provided a scalar value to perturb
                    # an indexed component.
                    ptb = value(perturbList[list_idx])

            # sipopt
            instance.sens_state_value_1[var] = ptb

            # k_aug
            # instance.DeltaP[con] = value(ptb - var)
            instance.DeltaP[con] = value(var - ptb)
            # FIXME: ^ This is incorrect. DeltaP should be (ptb - current).
            # But at least one test doesn't pass unless I use (current - ptb).



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.environ import AbstractModel, Var, Objective, minimize

model = AbstractModel()

model.x = Var()


def objective_rule(M):
    return M.x * M.x  # should fail "gracefully"


model.objective = Objective(rule=objective_rule, sense=minimize)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import pyomo.common.unittest as unittest

import pyomo.environ as pyo
from pyomo.core.expr.compare import compare_expressions
from pyomo.contrib.mpc.modeling.constraints import get_piecewise_constant_constraints


class TestPiecewiseConstantConstraints(unittest.TestCase):
    def _make_model(self, n_time_points=3):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=list(range(n_time_points)))
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(
            m.time, m.comp, initialize={(i, j): 1.1 * i for i, j in m.time * m.comp}
        )
        m.input = pyo.Var(m.time, initialize={i: 3.3 * i for i in m.time})
        return m

    def test_pwc_constraint_backward(self):
        # Tests the form of piecewise constant constraints that
        # should be used for a backward discretization, i.e.
        # that each input at non-sampling points is set equal to
        # the next value in the time set, rather than the previous.
        # This is the default.
        n_time_points = 5
        sample_points = [0, 2, 4]
        sample_points_set = set(sample_points)
        m = self._make_model(n_time_points=n_time_points)
        inputs = [pyo.Reference(m.var[:, "B"]), m.input]
        m.input_set, m.pwc_con = get_piecewise_constant_constraints(
            inputs, m.time, sample_points
        )
        pred_expr = {
            # Here we rely on knowledge that delta t == 1
            (i, t): inputs[i][t] - inputs[i][t + 1] == 0
            for t in m.time
            if t not in sample_points_set
            for i in range(len(inputs))
        }
        self.assertEqual(list(m.input_set), list(range(len(inputs))))
        for i in range(len(inputs)):
            for t in m.time:
                if t in sample_points_set:
                    self.assertNotIn((i, t), m.pwc_con)
                else:
                    self.assertIn((i, t), m.pwc_con)
                    self.assertEqual(
                        pyo.value(pred_expr[i, t]), pyo.value(m.pwc_con[i, t].expr)
                    )
                    self.assertTrue(
                        compare_expressions(pred_expr[i, t], m.pwc_con[i, t].expr)
                    )

    def test_pwc_constraint_forward(self):
        # The form of piecewise constant constraints that
        # should be used for a forward discretization.
        n_time_points = 5
        sample_points = [0, 2, 4]
        sample_points_set = set(sample_points)
        m = self._make_model(n_time_points=n_time_points)
        inputs = [pyo.Reference(m.var[:, "B"]), m.input]
        m.input_set, m.pwc_con = get_piecewise_constant_constraints(
            inputs, m.time, sample_points, use_next=False
        )
        pred_expr = {
            # Here we rely on knowledge that delta t == 1
            (i, t): inputs[i][t - 1] - inputs[i][t] == 0
            for t in m.time
            if t not in sample_points_set
            for i in range(len(inputs))
        }
        self.assertEqual(list(m.input_set), list(range(len(inputs))))
        for i in range(len(inputs)):
            for t in m.time:
                if t in sample_points_set:
                    self.assertNotIn((i, t), m.pwc_con)
                else:
                    self.assertIn((i, t), m.pwc_con)
                    self.assertEqual(
                        pyo.value(pred_expr[i, t]), pyo.value(m.pwc_con[i, t].expr)
                    )
                    self.assertTrue(
                        compare_expressions(pred_expr[i, t], m.pwc_con[i, t].expr)
                    )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Pyomo tutorials
#

import runpy
import sys
import os

import pyomo.common.unittest as unittest

from pyomo.common.fileutils import PYOMO_ROOT_DIR
from pyomo.common.tee import capture_output

tutorial_dir = os.path.join(PYOMO_ROOT_DIR, "examples", "pyomo", "tutorials")

try:
    from win32com.client.dynamic import Dispatch

    _win32com = True
except:
    _win32com = False  # pragma:nocover

from pyomo.common.dependencies import pyutilib, pyutilib_available

_excel_available = False
if _win32com and pyutilib_available:
    from pyutilib.excel.spreadsheet_win32com import ExcelSpreadsheet_win32com

    tmp = ExcelSpreadsheet_win32com()
    try:
        tmp._excel_dispatch()
        tmp._excel_quit()
        _excel_available = True
    except:
        pass

try:
    import xlrd

    _xlrd = True
except:
    _xlrd = False
try:
    import openpyxl

    _openpyxl = True
except:
    _openpyxl = False


class PyomoTutorials(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        self.tmp_path = list(sys.path)
        os.chdir(tutorial_dir)
        sys.path = [tutorial_dir] + sys.path

    def tearDown(self):
        os.chdir(self.cwd)
        sys.path = self.tmp_path

    def driver(self, name):
        with capture_output() as OUT:
            runpy.run_module(name, None, "__main__")
        self.assertIn(
            open(os.path.join(tutorial_dir, name + ".out"), 'r').read(), OUT.getvalue()
        )

    def test_data(self):
        self.driver('data')

    @unittest.skipUnless(_xlrd or _openpyxl, "Cannot read excel file.")
    @unittest.skipUnless(
        _win32com and _excel_available and pyutilib_available, "Cannot read excel file."
    )
    def test_excel(self):
        self.driver('excel')

    def test_set(self):
        self.driver('set')

    def test_table(self):
        self.driver('table')

    def test_param(self):
        self.driver('param')


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
from os.path import join, dirname, abspath
import types

try:
    import new

    new_available = True
except:
    new_available = False

import pyomo.common.unittest as unittest
from pyomo.opt import TerminationCondition
from pyomo.solvers.tests.models.base import all_models
from pyomo.solvers.tests.testcases import generate_scenarios
from pyomo.core.kernel.block import IBlock

# The test directory
thisDir = dirname(abspath(__file__))

# Cleanup Expected Failure Results Files
_cleanup_expected_failures = True


#
# A function that returns a function that gets
# added to a test class.
#
def create_method(test_name, model, solver, io, test_case, symbolic_labels):
    is_expected_failure = test_case.status == 'expected failure'

    #
    # Create a function that executes the test
    #
    def writer_test(self):
        # Create the model test class
        model_class = test_case.model()

        save_filename = join(thisDir, ("%s.soln.json" % model_class.description))
        # cleanup possibly existing old test files
        if os.path.exists(save_filename):
            os.remove(save_filename)

        # Create the model instance
        model_class.generate_model(test_case.testcase.import_suffixes)
        model_class.warmstart_model()

        # solve
        load_solutions = False
        opt, results = model_class.solve(
            solver,
            io,
            test_case.testcase.io_options,
            test_case.testcase.options,
            symbolic_labels,
            load_solutions,
        )
        termination_condition = results['Solver'][0]['termination condition']

        model_class.post_solve_test_validation(self, results)
        if (
            termination_condition == TerminationCondition.unbounded
            or termination_condition == TerminationCondition.infeasible
            or termination_condition == TerminationCondition.infeasibleOrUnbounded
        ):
            return

        # validate the solution returned by the solver
        if isinstance(model_class.model, IBlock):
            model_class.model.load_solution(results.Solution)
        else:
            model_class.model.solutions.load_from(
                results, default_variable_value=opt.default_variable_value()
            )
            model_class.save_current_solution(
                save_filename, suffixes=model_class.test_suffixes
            )
        rc = model_class.validate_current_solution(
            suffixes=model_class.test_suffixes,
            exclude_suffixes=test_case.exclude_suffixes,
        )

        if is_expected_failure:
            if rc[0]:
                self.fail(
                    "\nTest model '%s' was marked as an expected "
                    "failure but no failure occurred. The "
                    "reason given for the expected failure "
                    "is:\n\n****\n%s\n****\n\n"
                    "Please remove this case as an expected "
                    "failure if the above issue has been "
                    "corrected in the latest version of the "
                    "solver." % (model_class.description, test_case.msg)
                )
            if _cleanup_expected_failures:
                os.remove(save_filename)

        if not rc[0]:
            if not isinstance(model_class.model, IBlock):
                try:
                    model_class.model.solutions.store_to(results)
                except ValueError:
                    pass
            self.fail(
                "Solution mismatch for plugin "
                + test_name
                + ', '
                + io
                + " interface and problem type "
                + model_class.description
                + "\n"
                + rc[1]
                + "\n"
                + (str(results.Solution(0)) if len(results.solution) else "No Solution")
            )

        # cleanup if the test passed
        try:
            os.remove(save_filename)
        except OSError:
            pass

    # Skip this test if the status is 'skip'
    if test_case.status == 'skip':

        def return_test(self):
            return self.skipTest(test_case.msg)

    elif is_expected_failure:

        @unittest.expectedFailure
        def return_test(self):
            return writer_test(self)

    else:
        # Skip if solver is in demo mode
        size = getattr(test_case.model, 'size', (None, None, None))
        for prb, sol in zip(size, test_case.demo_limits):
            if (prb and sol) and prb > sol:

                def return_test(self):
                    return self.skipTest(
                        "Problem is too large for unlicensed %s solver" % solver
                    )

                break
            else:

                def return_test(self):
                    return writer_test(self)

    unittest.pytest.mark.solver(solver)(unittest.pytest.mark.writer(io)(return_test))
    return return_test


cls = None

#
# Create test driver classes for each test model
#
driver = {}
for model in all_models():
    # Get the test case for the model
    case = all_models(model)

    # Create the test class
    name = "Test_%s" % model
    if new_available:
        cls = new.classobj(name, (unittest.TestCase,), {})
    else:
        cls = types.new_class(name, (unittest.TestCase,))
        cls.__module__ = __name__
    driver[model] = cls
    globals()[name] = cls

#
# Iterate through all test scenarios and add test methods
#
for key, value in generate_scenarios():
    model, solver, io = key
    cls = driver[model]

    # Symbolic labels
    test_name = "test_" + solver + "_" + io + "_symbolic_labels"
    test_method = create_method(test_name, model, solver, io, value, True)
    if test_method is not None:
        setattr(cls, test_name, test_method)
        test_method = None

    # Non-symbolic labels
    test_name = "test_" + solver + "_" + io + "_nonsymbolic_labels"
    test_method = create_method(test_name, model, solver, io, value, False)
    if test_method is not None:
        setattr(cls, test_name, test_method)
        test_method = None

# Reset the cls variable, since it contains a unittest.TestCase subclass.
# This prevents this class from being processed twice!
cls = None

if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.environ import (
    ConcreteModel,
    RangeSet,
    Param,
    Var,
    Reals,
    Binary,
    Objective,
    Constraint,
    ConstraintList,
)
import math

N = 5
M = 6
P = 3

model = ConcreteModel()

model.Locations = RangeSet(1, N)

model.Customers = RangeSet(1, M)


def d_rule(model, n, m):
    return math.sin(n * 2.33333 + m * 7.99999)


model.d = Param(model.Locations, model.Customers, initialize=d_rule, within=Reals)

model.x = Var(model.Locations, model.Customers, bounds=(0.0, 1.0))

model.y = Var(model.Locations, within=Binary)


def rule(model):
    return sum(
        [
            model.d[n, m] * model.x[n, m]
            for n in model.Locations
            for m in model.Customers
        ]
    )


model.obj = Objective(rule=rule)

model.single_x = ConstraintList()
for m in model.Customers:
    model.single_x.add(sum([model.x[n, m] for n in model.Locations]) == 1.0)

model.bound_y = ConstraintList()
for n in model.Locations:
    for m in model.Customers:
        model.bound_y.add(model.x[n, m] <= model.y[n])

model.num_facilities = Constraint(expr=sum([model.y[n] for n in model.Locations]) == P)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
import sys
import copy
import json
import logging
import os.path

from pyomo.common.dependencies import yaml, yaml_load_args
import pyomo.opt
from pyomo.opt.results.container import undefined, ignore, ListContainer, MapContainer
import pyomo.opt.results.solution
from pyomo.opt.results.solution import default_print_options as dpo
import pyomo.opt.results.problem
import pyomo.opt.results.solver

from io import StringIO

logger = logging.getLogger(__name__)


def _guess_format(filename):
    "Return a standardized file format by looking at the filename extension"
    return {'.json': 'json', '.jsn': 'json', '.yaml': 'yaml', '.yml': 'yaml'}.get(
        os.path.splitext(filename)[1].lower(), None
    )


class SolverResults(MapContainer):
    undefined = undefined
    default_print_options = dpo

    def __init__(self):
        super().__init__()
        self._sections = []
        self._descriptions = {}
        self.add(
            'problem',
            ListContainer(pyomo.opt.results.problem.ProblemInformation),
            False,
            "Problem Information",
        )
        self.add(
            'solver',
            ListContainer(pyomo.opt.results.solver.SolverInformation),
            False,
            "Solver Information",
        )
        self.add(
            'solution',
            pyomo.opt.results.solution.SolutionSet(),
            False,
            "Solution Information",
        )

    def add(self, name, value, active, description):
        self.declare(name, value=value, active=active)
        tmp = self._convert(name)
        self._sections.append(tmp)
        self._descriptions[tmp] = description

    def json_repn(self, options=None):
        if options is None:
            return self._repn_(SolverResults.default_print_options)
        else:
            return self._repn_(options)

    def _repn_(self, option):
        if not option.schema and not self._active and not self._required:
            return ignore
        tmp = {}
        for key in self._sections:
            rep = dict.__getitem__(self, key)._repn_(option)
            if not rep == ignore:
                tmp[key] = rep
        return tmp

    def write(self, **kwds):
        _format = kwds.pop('format', None)
        if _format:
            _format = _format.lower()
        if 'filename' in kwds:
            filename = kwds.pop('filename')
            _guess = _guess_format(filename)
            if not _format:
                _format = _guess
            if _guess and _format != _guess:
                logger.warning(
                    "writing results to file (%s) using what appears "
                    "to be an incompatible format (%s)" % (filename, _format)
                )
            if 'ostream' in kwds:
                raise ValueError("Cannot specify both filename and ostream")
            with open(filename, "w") as OUTPUT:
                return self.write(ostream=OUTPUT, format=_format, **kwds)

        ostream = kwds.pop('ostream', sys.stdout)
        if not _format:
            _format = 'yaml'
        if _format == 'yaml':
            self.write_yaml(ostream=ostream, **kwds)
        elif _format == 'json':
            self.write_json(ostream=ostream, **kwds)
        else:
            raise ValueError("Unknown results file format: %s" % (_format,))

    def write_json(self, **kwds):
        ostream = kwds.pop('ostream', sys.stdout)
        option = copy.copy(SolverResults.default_print_options)
        # TODO: verify that we need this for-loop
        for key, val in kwds.items():
            setattr(option, key, val)
        repn = self.json_repn(option)

        for soln in repn.get('Solution', []):
            for data in ['Variable', 'Constraint', 'Objective']:
                remove = set()
                if data not in soln:
                    continue
                data_value = soln[data]
                if not isinstance(data_value, dict):
                    continue
                if not data_value:
                    # a variable/constraint/objective may have no
                    # entries, e.g., if duals or slacks weren't
                    # extracted in a solution.
                    soln[data] = "No values"
                    continue
                for kk, vv in data_value.items():
                    # TODO: remove this if-block.  This is a hack
                    if not type(vv) is dict:
                        vv = {'Value': vv}
                    tmp = {}
                    for k, v in vv.items():
                        # TODO: remove this if-block.  This is a hack
                        if v is not None and math.fabs(v) > 1e-16:
                            tmp[k] = v
                    if len(tmp) > 0:
                        soln[data][kk] = tmp
                    else:
                        remove.add((data, kk))
                for item in remove:
                    del soln[item[0]][item[1]]
        json.dump(repn, ostream, indent=4, sort_keys=True)

    def write_yaml(self, **kwds):
        ostream = kwds.pop('ostream', sys.stdout)
        option = copy.copy(SolverResults.default_print_options)
        # TODO: verify that we need this for-loop
        for key in kwds:
            setattr(option, key, kwds[key])
        repn = self._repn_(option)

        ostream.write("# ==========================================================\n")
        ostream.write("# = Solver Results                                         =\n")
        ostream.write("# ==========================================================\n")
        for key, item in self.items():
            if not key in repn:
                continue
            ostream.write(
                "# ----------------------------------------------------------\n"
            )
            ostream.write("#   %s\n" % self._descriptions[key])
            ostream.write(
                "# ----------------------------------------------------------\n"
            )
            ostream.write(key + ": ")
            if isinstance(item, ListContainer):
                item.pprint(ostream, option, prefix="", repn=repn[key])
            else:
                item.pprint(ostream, option, prefix="  ", repn=repn[key])

    def read(self, **kwds):
        _format = kwds.pop('format', None)
        if _format:
            _format = _format.lower()
        if 'filename' in kwds:
            filename = kwds.pop('filename')
            if not _format:
                _format = _guess_format(filename)
            if 'istream' in kwds:
                raise ValueError("Cannot specify both filename and istream")
            with open(filename, "r") as INPUT:
                return self.read(istream=INPUT, format=_format, **kwds)

        istream = kwds.pop('istream', sys.stdin)
        if not _format or _format == 'yaml':
            repn = yaml.load(istream, **yaml_load_args)
        elif _format == 'json':
            repn = json.load(istream)
        else:
            raise ValueError(f"Unknown SolverResults format: '{_format}'")
        for key, item in repn.items():
            dict.__getitem__(self, key).load(item)

    def __repr__(self):
        return str(self._repn_(SolverResults.default_print_options))

    def __str__(self):
        ostream = StringIO()
        option = SolverResults.default_print_options
        self.pprint(ostream, option, repn=self._repn_(option))
        return ostream.getvalue()


if __name__ == '__main__':
    results = SolverResults()
    results.write(schema=True)
    # print results



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import os
import re
import time
import logging
import subprocess

from pyomo.common import Executable
from pyomo.common.enums import maximize, minimize
from pyomo.common.errors import ApplicationError
from pyomo.common.tempfiles import TempfileManager

from pyomo.common.collections import ComponentMap, Bunch
from pyomo.opt.base import ProblemFormat, ResultsFormat, OptSolver, BranchDirection
from pyomo.opt.base.solvers import _extract_version, SolverFactory
from pyomo.opt.results import (
    SolverResults,
    SolverStatus,
    TerminationCondition,
    SolutionStatus,
    Solution,
)
from pyomo.opt.solver import ILMLicensedSystemCallSolver
from pyomo.solvers.mockmip import MockMIP
from pyomo.core.base import Var, Suffix, active_export_suffix_generator
from pyomo.core.kernel.suffix import export_suffix_generator
from pyomo.core.kernel.block import IBlock
from pyomo.util.components import iter_component

logger = logging.getLogger('pyomo.solvers')


def _validate_file_name(cplex, filename, description):
    """Validate filenames against the set of allowable characters in CPLEX.

    Returns the filename, possibly enclosed in double-quotes, or raises
    a ValueError is unallowable characters are found.

    """
    if filename is None:
        return filename
    matches = _validate_file_name.illegal_characters.search(filename)
    if matches:
        raise ValueError(
            "Unallowed character (%s) found in CPLEX %s file path/name.\n\t"
            "For portability reasons, only [%s] are allowed."
            % (
                matches.group(),
                description,
                _validate_file_name.allowed_characters.replace("\\", ''),
            )
        )
    # CPLEX only supports quoting spaces starting in v12.8.
    if ' ' in filename:
        if cplex.version()[:2] >= (12, 8):
            filename = '"' + filename + '"'
        else:
            raise ValueError(
                "Space detected in CPLEX %s file path/name\n\t%s\nand "
                "CPLEX older than version 12.8.  Please either upgrade "
                "CPLEX or remove the space from the %s path."
                % (description, filename, description)
            )
    return filename


# The full list of allowed characters, per IBM, is:
# (a-z, A-Z, 0-9) or ! " # $ % & ( ) / , . ; ? @ _ ` ' { } | ~
_validate_file_name.allowed_characters = (
    r"a-zA-Z0-9 ~:;,!'`|\$\(\)\{\}\?\#\&\.\-_\@\%s" % (os.path.sep,)
)
_validate_file_name.illegal_characters = re.compile(
    '[^%s]' % (_validate_file_name.allowed_characters,)
)


@SolverFactory.register('cplex', doc='The CPLEX LP/MIP solver')
class CPLEX(OptSolver):
    """The CPLEX LP/MIP solver"""

    def __new__(cls, *args, **kwds):
        try:
            mode = kwds['solver_io']
            if mode is None:
                mode = 'lp'
            del kwds['solver_io']
        except KeyError:
            mode = 'lp'
        #
        if mode == 'lp':
            return SolverFactory('_cplex_shell', **kwds)
        if mode == 'mps':
            opt = SolverFactory('_cplex_shell', **kwds)
            opt.set_problem_format(ProblemFormat.mps)
            return opt
        if mode in ['python', 'direct']:
            opt = SolverFactory('cplex_direct', **kwds)
            if opt is None:
                logging.getLogger('pyomo.solvers').error(
                    'Python API for CPLEX is not installed'
                )
                return
            return opt
        if mode == 'persistent':
            opt = SolverFactory('cplex_persistent', **kwds)
            if opt is None:
                logging.getLogger('pyomo.solvers').error(
                    'Python API for CPLEX is not installed'
                )
                return
            return opt
        #
        if mode == 'os':
            opt = SolverFactory('_ossolver', **kwds)
        elif mode == 'nl':
            opt = SolverFactory('asl', **kwds)
        else:
            logging.getLogger('pyomo.solvers').error('Unknown IO type: %s' % mode)
            return
        opt.set_options('solver=cplexamp')
        return opt


class ORDFileSchema:
    HEADER = "* ENCODING=ISO-8859-1\nNAME             Priority Order\n"
    FOOTER = "ENDATA\n"

    @classmethod
    def ROW(cls, name, priority, branch_direction=None):
        return " %s %s %s\n" % (cls._direction_to_str(branch_direction), name, priority)

    @staticmethod
    def _direction_to_str(branch_direction):
        try:
            return {BranchDirection.down: "DN", BranchDirection.up: "UP"}[
                branch_direction
            ]
        except KeyError:
            return ""


@SolverFactory.register(
    '_cplex_shell', doc='Shell interface to the CPLEX LP/MIP solver'
)
class CPLEXSHELL(ILMLicensedSystemCallSolver):
    """Shell interface to the CPLEX LP/MIP solver"""

    def __init__(self, **kwds):
        #
        # Call base class constructor
        #
        kwds['type'] = 'cplex'
        ILMLicensedSystemCallSolver.__init__(self, **kwds)

        # NOTE: eventually both of the following attributes should be migrated to a common base class.
        # is the current solve warm-started? a transient data member to communicate state information
        # across the _presolve, _apply_solver, and _postsolve methods.
        self._warm_start_solve = False
        # related to the above, the temporary name of the MST warm-start file (if any).
        self._warm_start_file_name = None

        #
        # Define valid problem formats and associated results formats
        #
        self._valid_problem_formats = [ProblemFormat.cpxlp, ProblemFormat.mps]
        self._valid_result_formats = {}
        self._valid_result_formats[ProblemFormat.cpxlp] = [ResultsFormat.soln]
        self._valid_result_formats[ProblemFormat.mps] = [ResultsFormat.soln]
        self.set_problem_format(ProblemFormat.cpxlp)

        # Note: Undefined capabilities default to 'None'
        self._capabilities = Bunch()
        self._capabilities.linear = True
        self._capabilities.quadratic_objective = True
        self._capabilities.quadratic_constraint = True
        self._capabilities.integer = True
        self._capabilities.sos1 = True
        self._capabilities.sos2 = True

    def _default_results_format(self, prob_format):
        return ResultsFormat.soln

    #
    # CPLEX has a simple, easy-to-use warm-start capability.
    #
    def warm_start_capable(self):
        return True

    #
    # write a warm-start file in the CPLEX MST format.
    #
    def _warm_start(self, instance):
        # for each variable in the symbol_map, add a child to the
        # variables element.  Both continuous and discrete are accepted
        # (and required, depending on other options), according to the
        # CPLEX manual.
        # **Note**: This assumes that the symbol_map is "clean", i.e.,
        # contains only references to the variables encountered in constraints
        output_index = 0
        if isinstance(instance, IBlock):
            smap = getattr(instance, "._symbol_maps")[self._smap_id]
        else:
            smap = instance.solutions.symbol_map[self._smap_id]
        byObject = smap.byObject

        # in principle, one could use a Python XML writer library like
        # xml.dom.minidom.  it works, but it is slow. hence, the
        # explicit direct-write of XML below.

        with open(self._warm_start_file_name, "w") as mst_file:
            mst_file.write("<?xml version=\"1.0\" ?>\n")
            mst_file.write("<CPLEXSolution version=\"1.0\">\n")
            mst_file.write("<header/>\n")
            mst_file.write("<quality/>\n")
            mst_file.write("<variables>\n")
            for var in instance.component_data_objects(Var):
                if (var.value is not None) and (id(var) in byObject):
                    name = byObject[id(var)]
                    mst_file.write(
                        "<variable index=\"%d\" "
                        "name=\"%s\" value=\"%f\" />\n"
                        % (output_index, name, var.value)
                    )
                    output_index = output_index + 1

            mst_file.write("</variables>\n")
            mst_file.write("</CPLEXSolution>\n")

    # Expected names of `Suffix` components for branching priorities and directions respectively
    SUFFIX_PRIORITY_NAME = "priority"
    SUFFIX_DIRECTION_NAME = "direction"

    def _write_priorities_file(self, instance):
        """Write a variable priorities file in the CPLEX ORD format."""
        priorities, directions = self._get_suffixes(instance)
        rows = self._convert_priorities_to_rows(instance, priorities, directions)
        self._write_priority_rows(rows)

    def _get_suffixes(self, instance):
        if isinstance(instance, IBlock):
            suffixes = (
                (suf.name, suf)
                for suf in export_suffix_generator(
                    instance, datatype=Suffix.INT, active=True, descend_into=False
                )
            )
        else:
            suffixes = active_export_suffix_generator(instance, datatype=Suffix.INT)
        suffixes = dict(suffixes)

        if self.SUFFIX_PRIORITY_NAME not in suffixes:
            raise ValueError(
                "Cannot write branching priorities file as `model.%s` Suffix has not been declared."
                % (self.SUFFIX_PRIORITY_NAME,)
            )

        return (
            suffixes[self.SUFFIX_PRIORITY_NAME],
            suffixes.get(self.SUFFIX_DIRECTION_NAME, ComponentMap()),
        )

    def _convert_priorities_to_rows(self, instance, priorities, directions):
        if isinstance(instance, IBlock):
            smap = getattr(instance, "._symbol_maps")[self._smap_id]
        else:
            smap = instance.solutions.symbol_map[self._smap_id]
        byObject = smap.byObject

        rows = []
        for var, priority in priorities.items():
            if priority is None or not var.active:
                continue

            if not (0 <= priority == int(priority)):
                raise ValueError("`priority` must be a non-negative integer")

            var_direction = directions.get(var, BranchDirection.default)

            for child_var in iter_component(var):
                if id(child_var) not in byObject:
                    continue

                child_var_direction = directions.get(child_var, var_direction)

                rows.append((byObject[id(child_var)], priority, child_var_direction))
        return rows

    def _write_priority_rows(self, rows):
        with open(self._priorities_file_name, "w") as ord_file:
            ord_file.write(ORDFileSchema.HEADER)
            for var_name, priority, direction in rows:
                ord_file.write(ORDFileSchema.ROW(var_name, priority, direction))
            ord_file.write(ORDFileSchema.FOOTER)

    # over-ride presolve to extract the warm-start keyword, if specified.
    def _presolve(self, *args, **kwds):
        # create a context in the temporary file manager for
        # this plugin - is "pop"ed in the _postsolve method.
        TempfileManager.push()

        # if the first argument is a string (representing a filename),
        # then we don't have an instance => the solver is being applied
        # to a file.
        self._warm_start_solve = kwds.pop('warmstart', False)
        self._warm_start_file_name = _validate_file_name(
            self, kwds.pop('warmstart_file', None), "warm start"
        )
        user_warmstart = self._warm_start_file_name is not None

        # the input argument can currently be one of two things: an instance or a filename.
        # if a filename is provided and a warm-start is indicated, we go ahead and
        # create the temporary file - assuming that the user has already, via some external
        # mechanism, invoked warm_start() with a instance to create the warm start file.
        if self._warm_start_solve and isinstance(args[0], str):
            # we assume the user knows what they are doing...
            pass
        elif self._warm_start_solve and (not isinstance(args[0], str)):
            # assign the name of the warm start file *before* calling the base class
            # presolve - the base class method ends up creating the command line,
            # and the warm start file-name is (obviously) needed there.
            if self._warm_start_file_name is None:
                assert not user_warmstart
                self._warm_start_file_name = TempfileManager.create_tempfile(
                    suffix='.cplex.mst'
                )

        self._priorities_solve = kwds.pop("priorities", False)
        self._priorities_file_name = _validate_file_name(
            self, kwds.pop("priorities_file", None), "branching priorities"
        )
        user_priorities = self._priorities_file_name is not None

        if (
            self._priorities_solve
            and not isinstance(args[0], str)
            and not user_priorities
        ):
            self._priorities_file_name = TempfileManager.create_tempfile(
                suffix=".cplex.ord"
            )

        # let the base class handle any remaining keywords/actions.
        ILMLicensedSystemCallSolver._presolve(self, *args, **kwds)

        # NB: we must let the base class presolve run first so that the
        # symbol_map is actually constructed!

        if (len(args) > 0) and (not isinstance(args[0], str)):
            if len(args) != 1:
                raise ValueError(
                    "CPLEX _presolve method can only handle a "
                    "single problem instance - %s were supplied" % (len(args),)
                )

            # write the warm-start file - currently only supports MIPs.
            # we only know how to deal with a single problem instance.
            if self._warm_start_solve and (not user_warmstart):
                start_time = time.time()
                self._warm_start(args[0])
                end_time = time.time()
                if self._report_timing:
                    print(
                        "Warm start write time= %.2f seconds" % (end_time - start_time)
                    )

            if self._priorities_solve and (not user_priorities):
                start_time = time.time()
                self._write_priorities_file(args[0])
                end_time = time.time()
                if self._report_timing:
                    print(
                        "Branching priorities write time= %.2f seconds"
                        % (end_time - start_time)
                    )

    def _default_executable(self):
        executable = Executable("cplex")
        if not executable:
            logger.warning(
                "Could not locate the 'cplex' executable"
                ", which is required for solver %s" % self.name
            )
            self.enable = False
            return None
        return executable.path()

    def _get_version(self):
        """
        Returns a tuple describing the solver executable version.
        """
        solver_exec = self.executable()
        if solver_exec is None:
            return _extract_version('')
        results = subprocess.run(
            [solver_exec, '-c', 'quit'],
            timeout=self._version_timeout,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        return _extract_version(results.stdout)

    def create_command_line(self, executable, problem_files):
        #
        # Define log file
        # The log file in CPLEX contains the solution trace, but the solver status can be found in the solution file.
        #
        if self._log_file is None:
            self._log_file = TempfileManager.create_tempfile(suffix='.cplex.log')
        self._log_file = _validate_file_name(self, self._log_file, "log")

        #
        # Define solution file
        # As indicated above, contains (in XML) both the solution and solver status.
        #
        if self._soln_file is None:
            self._soln_file = TempfileManager.create_tempfile(suffix='.cplex.sol')
        self._soln_file = _validate_file_name(self, self._soln_file, "solution")

        #
        # Write the CPLEX execution script
        #
        script = 'set logfile %s\n' % (self._log_file,)
        if self._timelimit is not None and self._timelimit > 0.0:
            script += 'set timelimit %s\n' % (self._timelimit,)

        if (self.options.mipgap is not None) and (float(self.options.mipgap) > 0.0):
            script += 'set mip tolerances mipgap %s\n' % (self.options.mipgap,)
        for key in self.options:
            if key == 'relax_integrality' or key == 'mipgap':
                continue
            elif isinstance(self.options[key], str) and (' ' in self.options[key]):
                opt = ' '.join(key.split('_')) + ' ' + str(self.options[key])
            else:
                opt = ' '.join(key.split('_')) + ' ' + str(self.options[key])
            script += 'set %s\n' % (opt,)

        _lp_file = _validate_file_name(self, problem_files[0], "LP")
        script += 'read %s\n' % (_lp_file,)

        # if we're dealing with an LP, the MST file will be empty.
        if self._warm_start_solve and (self._warm_start_file_name is not None):
            script += 'read %s\n' % (self._warm_start_file_name,)

        if self._priorities_solve and self._priorities_file_name is not None:
            script += "read %s\n" % (self._priorities_file_name,)

        if 'relax_integrality' in self.options:
            script += 'change problem lp\n'

        script += 'display problem stats\n'
        script += 'optimize\n'
        script += 'write %s\n' % (self._soln_file,)
        script += 'quit\n'

        # dump the script and warm-start file names for the
        # user if we're keeping files around.
        if self._keepfiles:
            script_fname = TempfileManager.create_tempfile(suffix='.cplex.script')
            tmp = open(script_fname, 'w')
            tmp.write(script)
            tmp.close()

            print("Solver script file=" + script_fname)
            if self._warm_start_solve and (self._warm_start_file_name is not None):
                print("Solver warm-start file=" + self._warm_start_file_name)

            if self._priorities_solve and self._priorities_file_name is not None:
                print("Solver priorities file=" + self._priorities_file_name)

        #
        # Define command line
        #
        cmd = [executable]
        if self._timer:
            cmd.insert(0, self._timer)
        return Bunch(cmd=cmd, script=script, log_file=self._log_file, env=None)

    def process_logfile(self):
        """
        Process logfile
        """
        results = SolverResults()
        results.problem.number_of_variables = None
        results.problem.number_of_nonzeros = None
        #
        # Process logfile
        #
        OUTPUT = open(self._log_file)
        output = "".join(OUTPUT.readlines())
        OUTPUT.close()
        #
        # It is generally useful to know the CPLEX version number for logfile parsing.
        #
        cplex_version = None

        #
        # Parse logfile lines
        #

        # caching for subsequent use - we need to known the problem sense before using this information.
        # adding to plugin to cache across invocation of process_logfile and process_soln_file.
        self._best_bound = None
        self._gap = None

        for line in output.split("\n"):
            tokens = re.split('[ \t]+', line.strip())
            if len(tokens) > 3 and tokens[0] == "CPLEX" and tokens[1] == "Error":
                # IMPT: See below - cplex can generate an error line and then terminate fine, e.g., in CPLEX 12.1.
                #       To handle these cases, we should be specifying some kind of termination criterion always
                #       in the course of parsing a log file (we aren't doing so currently - just in some conditions).
                results.solver.status = SolverStatus.error
                results.solver.error = " ".join(tokens)
            elif len(tokens) >= 3 and tokens[0] == "ILOG" and tokens[1] == "CPLEX":
                cplex_version = tokens[2].rstrip(',')
            elif len(tokens) >= 3 and tokens[0] == "Variables":
                if (
                    results.problem.number_of_variables is None
                ):  # CPLEX 11.2 and subsequent versions have two Variables sections in the log file output.
                    results.problem.number_of_variables = int(tokens[2])
            # In CPLEX 11 (and presumably before), there was only a single line output to
            # indicate the constraint count, e.g., "Linear constraints : 16 [Less: 7, Greater: 6, Equal: 3]".
            # In CPLEX 11.2 (or somewhere in between 11 and 11.2 - I haven't bothered to track it down
            # in that detail), there is another instance of this line prefix in the min/max problem statistics
            # block - which we don't care about. In this case, the line looks like: "Linear constraints :" and
            # that's all.
            elif (
                len(tokens) >= 4
                and tokens[0] == "Linear"
                and tokens[1] == "constraints"
            ):
                results.problem.number_of_constraints = int(tokens[3])
            elif len(tokens) >= 3 and tokens[0] == "Nonzeros":
                if (
                    results.problem.number_of_nonzeros is None
                ):  # CPLEX 11.2 and subsequent has two Nonzeros sections.
                    results.problem.number_of_nonzeros = int(tokens[2])
            elif len(tokens) >= 5 and tokens[4] == "MINIMIZE":
                results.problem.sense = minimize
            elif len(tokens) >= 5 and tokens[4] == "MAXIMIZE":
                results.problem.sense = maximize
            elif (
                len(tokens) >= 4
                and tokens[0] == "Solution"
                and tokens[1] == "time"
                and tokens[2] == "="
            ):
                # technically, I'm not sure if this is CPLEX user time or user+system - CPLEX doesn't appear
                # to differentiate, and I'm not sure we can always provide a break-down.
                results.solver.user_time = float(tokens[3])
            elif (
                len(tokens) >= 4
                and tokens[0] == "Primal"
                and tokens[1] == "simplex"
                and tokens[3] == "Optimal:"
            ):
                results.solver.termination_condition = TerminationCondition.optimal
                results.solver.termination_message = ' '.join(tokens)
            elif (
                len(tokens) >= 4
                and tokens[0] == "Dual"
                and tokens[1] == "simplex"
                and tokens[3] == "Optimal:"
            ):
                results.solver.termination_condition = TerminationCondition.optimal
                results.solver.termination_message = ' '.join(tokens)
            elif (
                len(tokens) >= 4 and tokens[0] == "Barrier" and tokens[2] == "Optimal:"
            ):
                results.solver.termination_condition = TerminationCondition.optimal
                results.solver.termination_message = ' '.join(tokens)
            elif (
                len(tokens) >= 4 and tokens[0] == "Dual" and tokens[3] == "Infeasible:"
            ):
                results.solver.termination_condition = TerminationCondition.infeasible
                results.solver.termination_message = ' '.join(tokens)
            elif (
                len(tokens) >= 4
                and tokens[0] == "MIP"
                and tokens[2] == "Integer"
                and tokens[3] == "infeasible."
            ):
                # if CPLEX has previously printed an error message, reduce it to a warning -
                # there is a strong indication it recovered, but we can't be sure.
                if results.solver.status == SolverStatus.error:
                    results.solver.status = SolverStatus.warning
                else:
                    results.solver.status = SolverStatus.ok
                results.solver.termination_condition = TerminationCondition.infeasible
                results.solver.termination_message = ' '.join(tokens)
            elif (
                len(tokens) >= 10
                and tokens[0] == "MIP"
                and tokens[2] == "Time"
                and tokens[3] == "limit"
                and tokens[6] == "feasible:"
            ):
                # handle processing when the time limit has been exceeded, and we have a feasible solution.
                results.solver.status = SolverStatus.ok
                results.solver.termination_condition = TerminationCondition.maxTimeLimit
                results.solver.termination_message = ' '.join(tokens)
            elif (
                len(tokens) >= 10
                and tokens[0] == "Current"
                and tokens[1] == "MIP"
                and tokens[2] == "best"
                and tokens[3] == "bound"
            ):
                self._best_bound = float(tokens[5])
                self._gap = float(tokens[8].rstrip(','))
            # for the case below, CPLEX sometimes reports "true" optimal (the first case)
            # and other times within-tolerance optimal (the second case).
            elif (
                len(tokens) >= 4
                and tokens[0] == "MIP"
                and tokens[2] == "Integer"
                and tokens[3] == "optimal"
            ) or (
                len(tokens) >= 4
                and tokens[0] == "MIP"
                and tokens[2] == "Integer"
                and tokens[3] == "optimal,"
            ):
                # if CPLEX has previously printed an error message, reduce it to a warning -
                # there is a strong indication it recovered, but we can't be sure.
                if results.solver.status == SolverStatus.error:
                    results.solver.status = SolverStatus.warning
                else:
                    results.solver.status = SolverStatus.ok
                results.solver.termination_condition = TerminationCondition.optimal
                results.solver.termination_message = ' '.join(tokens)
            elif (
                len(tokens) >= 3
                and tokens[0] == "Presolve"
                and tokens[2] == "Infeasible."
            ):
                # if CPLEX has previously printed an error message, reduce it to a warning -
                # there is a strong indication it recovered, but we can't be sure.
                if results.solver.status == SolverStatus.error:
                    results.solver.status = SolverStatus.warning
                else:
                    results.solver.status = SolverStatus.ok
                results.solver.termination_condition = TerminationCondition.infeasible
                results.solver.termination_message = ' '.join(tokens)
            elif (
                (
                    (len(tokens) == 6)
                    and (tokens[2] == "Integer")
                    and (tokens[3] == "infeasible")
                    and (tokens[5] == "unbounded.")
                )
                or (
                    (len(tokens) >= 4)
                    and (tokens[0] == "MIP")
                    and (tokens[1] == "-")
                    and (tokens[2] == "Integer")
                    and (tokens[3] == "unbounded:")
                )
                or (
                    (len(tokens) >= 5)
                    and (tokens[0] == "Presolve")
                    and (tokens[2] == "Unbounded")
                    and (tokens[4] == "infeasible.")
                )
            ):
                # if CPLEX has previously printed an error message, reduce it to a warning -
                # there is a strong indication it recovered, but we can't be sure.
                if results.solver.status == SolverStatus.error:
                    results.solver.status = SolverStatus.warning
                else:
                    results.solver.status = SolverStatus.ok
                # It isn't clear whether we can determine if the problem is unbounded from
                # CPLEX's output.
                results.solver.termination_condition = TerminationCondition.unbounded
                results.solver.termination_message = ' '.join(tokens)

        try:
            if isinstance(results.solver.termination_message, str):
                results.solver.termination_message = (
                    results.solver.termination_message.replace(':', '\\x3a')
                )
        except:
            pass
        return results

    def process_soln_file(self, results):
        # the only suffixes that we extract from CPLEX are
        # constraint duals, constraint slacks, and variable
        # reduced-costs. scan through the solver suffix list
        # and throw an exception if the user has specified
        # any others.
        extract_duals = False
        extract_slacks = False
        extract_reduced_costs = False
        extract_rc = False
        extract_lrc = False
        extract_urc = False
        for suffix in self._suffixes:
            flag = False
            if re.match(suffix, "dual"):
                extract_duals = True
                flag = True
            if re.match(suffix, "slack"):
                extract_slacks = True
                flag = True
            if re.match(suffix, "rc"):
                extract_reduced_costs = True
                extract_rc = True
                flag = True
            if re.match(suffix, "lrc"):
                extract_reduced_costs = True
                extract_lrc = True
                flag = True
            if re.match(suffix, "urc"):
                extract_reduced_costs = True
                extract_urc = True
                flag = True
            if not flag:
                raise RuntimeError(
                    "***The CPLEX solver plugin cannot extract solution suffix="
                    + suffix
                )

        # check for existence of the solution file
        # not sure why we just return - would think that we
        # would want to indicate some sort of error
        if not os.path.exists(self._soln_file):
            return

        range_duals = {}
        range_slacks = {}
        soln = Solution()
        soln.objective['__default_objective__'] = {'Value': None}

        # caching for efficiency
        soln_variables = soln.variable
        soln_constraints = soln.constraint

        INPUT = open(self._soln_file, "r")
        results.problem.number_of_objectives = 1
        time_limit_exceeded = False
        mip_problem = False
        for line in INPUT:
            line = line.strip()
            line = line.lstrip('<?/')
            line = line.rstrip('/>?')
            tokens = line.split(' ')

            if tokens[0] == "variable":
                variable_name = None
                variable_value = None
                variable_reduced_cost = None
                variable_status = None
                for i in range(1, len(tokens)):
                    field_name = tokens[i].split('=')[0]
                    field_value = tokens[i].split('=')[1].lstrip("\"").rstrip("\"")
                    if field_name == "name":
                        variable_name = field_value
                    elif field_name == "value":
                        variable_value = field_value
                    elif (extract_reduced_costs is True) and (
                        field_name == "reducedCost"
                    ):
                        variable_reduced_cost = field_value
                    elif (extract_reduced_costs is True) and (field_name == "status"):
                        variable_status = field_value

                # skip the "constant-one" variable, used to capture/retain objective offsets in the CPLEX LP format.
                if variable_name != "ONE_VAR_CONSTANT":
                    variable = soln_variables[variable_name] = {
                        "Value": float(variable_value)
                    }
                    if (variable_reduced_cost is not None) and (
                        extract_reduced_costs is True
                    ):
                        try:
                            if extract_rc is True:
                                variable["Rc"] = float(variable_reduced_cost)
                            if variable_status is not None:
                                if extract_lrc is True:
                                    if variable_status == "LL":
                                        variable["Lrc"] = float(variable_reduced_cost)
                                    else:
                                        variable["Lrc"] = 0.0
                                if extract_urc is True:
                                    if variable_status == "UL":
                                        variable["Urc"] = float(variable_reduced_cost)
                                    else:
                                        variable["Urc"] = 0.0
                        except:
                            raise ValueError(
                                "Unexpected reduced-cost value="
                                + str(variable_reduced_cost)
                                + " encountered for variable="
                                + variable_name
                            )
            elif (tokens[0] == "constraint") and (
                (extract_duals is True) or (extract_slacks is True)
            ):
                is_range = False
                rlabel = None
                rkey = None
                for i in range(1, len(tokens)):
                    field_name = tokens[i].split('=')[0]
                    field_value = tokens[i].split('=')[1].lstrip("\"").rstrip("\"")
                    if field_name == "name":
                        if field_value.startswith('c_'):
                            constraint = soln_constraints[field_value] = {}
                        elif field_value.startswith('r_l_'):
                            is_range = True
                            rlabel = field_value[4:]
                            rkey = 0
                        elif field_value.startswith('r_u_'):
                            is_range = True
                            rlabel = field_value[4:]
                            rkey = 1
                    elif (extract_duals is True) and (field_name == "dual"):  # for LPs
                        if is_range is False:
                            constraint["Dual"] = float(field_value)
                        else:
                            range_duals.setdefault(rlabel, [0, 0])[rkey] = float(
                                field_value
                            )
                    elif (extract_slacks is True) and (
                        field_name == "slack"
                    ):  # for MIPs
                        if is_range is False:
                            constraint["Slack"] = float(field_value)
                        else:
                            range_slacks.setdefault(rlabel, [0, 0])[rkey] = float(
                                field_value
                            )
            elif tokens[0].startswith("problemName"):
                filename = (tokens[0].split('=')[1].strip()).lstrip("\"").rstrip("\"")
                results.problem.name = os.path.basename(filename)
                if '.' in results.problem.name:
                    results.problem.name = results.problem.name.split('.')[0]
                tINPUT = open(filename, "r")
                for tline in tINPUT:
                    tline = tline.strip()
                    if tline == "":
                        continue
                    tokens = re.split('[\t ]+', tline)
                    if tokens[0][0] in ['\\', '*']:
                        continue
                    elif tokens[0] == "NAME":
                        results.problem.name = tokens[1]
                    else:
                        sense = tokens[0].lower()
                        if sense in ['max', 'maximize']:
                            results.problem.sense = maximize
                        if sense in ['min', 'minimize']:
                            results.problem.sense = minimize
                    break
                tINPUT.close()

            elif (
                tokens[0].startswith("objectiveValue")
                and tokens[0] != 'objectiveValues'
            ):
                # prior to 12.10.0, the objective value came back as an
                # attribute on the <header> tag
                objective_value = (
                    (tokens[0].split('=')[1].strip()).lstrip("\"").rstrip("\"")
                )
                soln.objective['__default_objective__']['Value'] = float(
                    objective_value
                )

            elif tokens[0] == "objective":
                # beginning in 12.10.0, CPLEX supports multiple
                # objectives in an <objectiveValue> tag
                fields = {}
                for field in tokens[1:]:
                    k, v = field.split('=')
                    fields[k] = v.strip('"')
                soln.objective.setdefault(fields['name'], {})['Value'] = float(
                    fields['value']
                )

            elif tokens[0].startswith("solutionStatusValue"):
                pieces = tokens[0].split("=")
                solution_status = eval(pieces[1])
                # solution status = 1 => optimal
                # solution status = 3 => infeasible
                if soln.status == SolutionStatus.unknown:
                    if solution_status == 1:
                        soln.status = SolutionStatus.optimal
                    elif solution_status == 3:
                        soln.status = SolutionStatus.infeasible
                        soln.gap = None
                    else:
                        # we are flagging anything with a solution status >= 4 as an error, to possibly
                        # be over-ridden as we learn more about the status (e.g., due to time limit exceeded).
                        soln.status = SolutionStatus.error
                        soln.gap = None
            elif tokens[0].startswith("solutionStatusString"):
                solution_status = (
                    ((" ".join(tokens).split('=')[1]).strip()).lstrip("\"").rstrip("\"")
                )
                if solution_status in [
                    "optimal",
                    "integer optimal solution",
                    "integer optimal, tolerance",
                ]:
                    soln.status = SolutionStatus.optimal
                    soln.gap = 0.0
                    results.problem.lower_bound = soln.objective[
                        '__default_objective__'
                    ]['Value']
                    results.problem.upper_bound = soln.objective[
                        '__default_objective__'
                    ]['Value']
                    if "integer" in solution_status:
                        mip_problem = True
                elif solution_status in ["infeasible"]:
                    soln.status = SolutionStatus.infeasible
                    soln.gap = None
                elif solution_status in ["time limit exceeded"]:
                    # we need to know if the solution is primal feasible, and if it is, set the solution status accordingly.
                    # for now, just set the flag so we can trigger the logic when we see the primalFeasible keyword.
                    time_limit_exceeded = True
            elif tokens[0].startswith("MIPNodes"):
                if mip_problem:
                    n = eval(
                        eval((" ".join(tokens).split('=')[1]).strip())
                        .lstrip("\"")
                        .rstrip("\"")
                    )
                    results.solver.statistics.branch_and_bound.number_of_created_subproblems = (
                        n
                    )
                    results.solver.statistics.branch_and_bound.number_of_bounded_subproblems = (
                        n
                    )
            elif tokens[0].startswith("primalFeasible") and (
                time_limit_exceeded is True
            ):
                primal_feasible = int(
                    ((" ".join(tokens).split('=')[1]).strip()).lstrip("\"").rstrip("\"")
                )
                if primal_feasible == 1:
                    soln.status = SolutionStatus.feasible
                    if results.problem.sense == minimize:
                        results.problem.upper_bound = soln.objective[
                            '__default_objective__'
                        ]['Value']
                    else:
                        results.problem.lower_bound = soln.objective[
                            '__default_objective__'
                        ]['Value']
                else:
                    soln.status = SolutionStatus.infeasible

        if self._best_bound is not None:
            if results.problem.sense == minimize:
                results.problem.lower_bound = self._best_bound
            else:
                results.problem.upper_bound = self._best_bound
        if self._gap is not None:
            soln.gap = self._gap

        # For the range constraints, supply only the dual with the largest
        # magnitude (at least one should always be numerically zero)
        for key, (ld, ud) in range_duals.items():
            if abs(ld) > abs(ud):
                soln_constraints['r_l_' + key] = {"Dual": ld}
            else:
                soln_constraints['r_l_' + key] = {"Dual": ud}  # Use the same key
        # slacks
        for key, (ls, us) in range_slacks.items():
            if abs(ls) > abs(us):
                soln_constraints.setdefault('r_l_' + key, {})["Slack"] = ls
            else:
                soln_constraints.setdefault('r_l_' + key, {})[
                    "Slack"
                ] = us  # Use the same key

        if not results.solver.status is SolverStatus.error:
            if results.solver.termination_condition in [
                TerminationCondition.unknown,
                # TerminationCondition.maxIterations,
                # TerminationCondition.minFunctionValue,
                # TerminationCondition.minStepLength,
                TerminationCondition.globallyOptimal,
                TerminationCondition.locallyOptimal,
                TerminationCondition.optimal,
                # TerminationCondition.maxEvaluations,
                TerminationCondition.other,
            ]:
                results.solution.insert(soln)
            elif (
                results.solver.termination_condition
                is TerminationCondition.maxTimeLimit
            ) and (soln.status is not SolutionStatus.infeasible):
                results.solution.insert(soln)

        INPUT.close()

    def _postsolve(self):
        # take care of the annoying (and empty) CPLEX temporary files in the current directory.
        # this approach doesn't seem overly efficient, but python os module functions don't
        # accept regular expression directly.
        filename_list = os.listdir(".")
        for filename in filename_list:
            # CPLEX temporary files come in two flavors - cplex.log and clone*.log.
            # the latter is the case for multi-processor environments.
            # IMPT: trap the possible exception raised by the file not existing.
            #       this can occur in pyro environments where > 1 workers are
            #       running CPLEX, and were started from the same directory.
            #       these logs don't matter anyway (we redirect everything),
            #       and are largely an annoyance.
            try:
                if re.match(r'cplex\.log', filename) != None:
                    os.remove(filename)
                elif re.match(r'clone\d+\.log', filename) != None:
                    os.remove(filename)
            except OSError:
                pass

        # let the base class deal with returning results.
        results = ILMLicensedSystemCallSolver._postsolve(self)

        # finally, clean any temporary files registered with the temp file
        # manager, created populated *directly* by this plugin. does not
        # include, for example, the execution script. but does include
        # the warm-start file.
        TempfileManager.pop(remove=not self._keepfiles)

        return results


@SolverFactory.register('_mock_cplex')
class MockCPLEX(CPLEXSHELL, MockMIP):
    """A Mock CPLEX solver used for testing"""

    def __init__(self, **kwds):
        try:
            CPLEXSHELL.__init__(self, **kwds)
        except ApplicationError:  # pragma:nocover
            pass  # pragma:nocover
        MockMIP.__init__(self, "cplex")

    def available(self, exception_flag=True):
        return CPLEXSHELL.available(self, exception_flag)

    def create_command_line(self, executable, problem_files):
        command = CPLEXSHELL.create_command_line(self, executable, problem_files)
        MockMIP.create_command_line(self, executable, problem_files)
        return command

    def _default_executable(self):
        return MockMIP.executable(self)

    def _execute_command(self, cmd):
        return MockMIP._execute_command(self, cmd)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Transformation to reformulate integer variables into binary."""

from math import floor, log
import logging

from pyomo.common.collections import ComponentSet
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.core import (
    TransformationFactory,
    Var,
    Block,
    Constraint,
    Any,
    Binary,
    value,
    RangeSet,
    Reals,
)
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.gdp import Disjunct
from pyomo.core.expr import identify_variables
from pyomo.common.modeling import unique_component_name

logger = logging.getLogger('pyomo.contrib.preprocessing')


@TransformationFactory.register(
    'contrib.integer_to_binary',
    doc="Reformulate integer variables into binary variables.",
)
class IntegerToBinary(IsomorphicTransformation):
    """Reformulate integer variables to binary variables and constraints.

    This transformation may be safely applied multiple times to the same model.
    """

    CONFIG = ConfigBlock("contrib.integer_to_binary")
    CONFIG.declare(
        "strategy",
        ConfigValue(
            default='base2',
            domain=In('base2'),
            description="Reformulation method",
            # TODO: eventually we will support other methods, but not yet.
        ),
    )
    CONFIG.declare(
        "ignore_unused",
        ConfigValue(
            default=False,
            domain=bool,
            description="Ignore variables that do not appear in (potentially) active constraints. "
            "These variables are unlikely to be passed to the solver.",
        ),
    )
    CONFIG.declare(
        "relax_integrality",
        ConfigValue(
            default=True,
            domain=bool,
            description="Relax the integrality of the integer variables "
            "after adding in the binary variables and constraints.",
        ),
    )

    def _apply_to(self, model, **kwds):
        """Apply the transformation to the given model."""
        config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)

        integer_vars = list(
            v
            for v in model.component_data_objects(
                ctype=Var, descend_into=(Block, Disjunct)
            )
            if v.is_integer() and not v.is_binary() and not v.fixed
        )
        if len(integer_vars) == 0:
            logger.info("Model has no free integer variables. No reformulation needed.")
            return

        vars_on_constr = ComponentSet()
        for c in model.component_data_objects(
            ctype=Constraint, descend_into=(Block, Disjunct), active=True
        ):
            vars_on_constr.update(
                v
                for v in identify_variables(c.body, include_fixed=False)
                if v.is_integer()
            )

        if config.ignore_unused:
            num_vars_not_on_constr = len(integer_vars) - len(vars_on_constr)
            if num_vars_not_on_constr > 0:
                logger.info(
                    "%s integer variables on the model are not attached to any constraints. "
                    "Ignoring unused variables."
                )
            integer_vars = list(vars_on_constr)

        logger.info(
            "Reformulating integer variables using the %s strategy." % config.strategy
        )

        # Set up reformulation block
        blk_name = unique_component_name(model, "_int_to_binary_reform")
        reform_block = Block(
            doc="Holds variables and constraints for reformulating "
            "integer variables to binary variables."
        )
        setattr(model, blk_name, reform_block)

        reform_block.int_var_set = RangeSet(0, len(integer_vars) - 1)

        reform_block.new_binary_var = Var(
            Any,
            domain=Binary,
            dense=False,
            initialize=0,
            doc="Binary variable with index (int_var_idx, idx)",
        )
        reform_block.integer_to_binary_constraint = Constraint(
            reform_block.int_var_set,
            doc="Equality constraints mapping the binary variable values "
            "to the integer variable value.",
        )

        # check that variables are bounded
        for idx, int_var in enumerate(integer_vars):
            if not (int_var.has_lb() and int_var.has_ub()):
                raise ValueError(
                    "Integer variable %s is missing an "
                    "upper or lower bound. LB: %s; UB: %s. "
                    "Integer to binary reformulation does not support unbounded integer variables."
                    % (int_var.name, int_var.lb, int_var.ub)
                )
            # do the reformulation
            highest_power = int(floor(log(value(int_var.ub - int_var.lb), 2)))
            # TODO potentially fragile due to floating point

            reform_block.integer_to_binary_constraint.add(
                idx,
                expr=int_var
                == sum(
                    reform_block.new_binary_var[idx, pwr] * (2**pwr)
                    for pwr in range(0, highest_power + 1)
                )
                + int_var.lb,
            )

            # Relax the original integer variable
            if config.relax_integrality:
                int_var.domain = Reals

        logger.info(
            "Reformulated %s integer variables using "
            "%s binary variables and %s constraints."
            % (
                len(integer_vars),
                len(reform_block.new_binary_var),
                len(reform_block.integer_to_binary_constraint),
            )
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
from weakref import ref as weakref_ref

from pyomo.common.log import is_debug_set
from pyomo.core.base.set_types import Any
from pyomo.core.base.var import IndexedVar, VarData
from pyomo.core.base.constraint import IndexedConstraint, ConstraintData
from pyomo.core.base.objective import IndexedObjective, ObjectiveData
from pyomo.core.base.expression import IndexedExpression, ExpressionData

from collections.abc import MutableSequence

logger = logging.getLogger('pyomo.core')

#
# In the future I think ComponentDict and ComponentList should inherit
# directly from (Active)IndexedComponent, and that class should be
# stripped down to a minimal interface. The abstract interface should
# be implemented on top of these classes.
#


class ComponentList(MutableSequence):
    def __init__(self, interface_datatype, *args):
        self._interface_datatype = interface_datatype
        self._data = []
        if len(args) > 0:
            if len(args) > 1:
                raise TypeError(
                    "ComponentList expected at most 1 arguments, "
                    "got %s" % (len(args))
                )
            for item in args[0]:
                self.append(item)

    def construct(self, data=None):
        if is_debug_set(logger):
            logger.debug(  # pragma:nocover
                "Constructing ComponentList object, name=%s, from data=%s"
                % (self.name, str(data))
            )
        if self._constructed:  # pragma:nocover
            return
        self._constructed = True

    #
    # Note: The following methods add a flavor of the dict-interface
    #       to this component. This is simply a hack that allows us to
    #       create working Pyomo examples using this prototype. The
    #       correct way to go would be to remove these methods and to
    #       update pyomo code everywhere to first check whether it is
    #       dealing with a dict or a list interface before
    #       iterating. I don't think that would be difficult to do.
    #

    def keys(self):
        return range(len(self))

    iterkeys = keys

    def values(self):
        return list(iter(self))

    itervalues = values

    def items(self):
        return zip(self.keys(), self.values())

    iteritems = items

    #
    # Define the MutableSequence abstract methods
    #

    # Currently __setitem__ only supports assignment of an explicit
    # instantiation of the interface datatype. Updates to an already
    # existing object inside this container need to be done via that
    # objects interface methods or by directly accessing object
    # attributes.
    #
    # * See notes above the __setitem__ method for ComponentDict
    #   about potentially supporting implicit assignment / update
    #
    def __setitem__(self, i, item):
        if isinstance(item, self._interface_datatype):
            # release the current component (assuming we don't get
            # an index error)
            # * see __delitem__ for explanation
            if item._component is None:
                item._component = weakref_ref(self)
                if hasattr(self, "_active"):
                    self._active |= getattr(item, '_active', True)
                self._data[i]._component = None
                self._data[i] = item
                item._index = i
                return
            elif self._data[i] is item:
                # a very special case that makes sense to handle
                # because the implied order should be: (1) delete
                # the object at the current index, (2) insert the
                # the new object. This performs both without any
                # actions, but it is an extremely rare case, so
                # it should go last.
                return
            # see note about allowing components to live in more than
            # one container
            raise ValueError(
                "Invalid component object assignment to ComponentList "
                "%s at index %s. A parent component has already been "
                "assigned the object: %s" % (self.name, i, item.parent_component().name)
            )
        # see note about implicit assignment and update
        raise TypeError(
            "ComponentList must be assigned objects "
            "of type %s. Invalid type for key %s: %s"
            % (self._interface_datatype.__name__, i, type(item))
        )

    # * Only supports explicit objects. See notes above __setitem__
    #   for more information
    def insert(self, i, item):
        if isinstance(item, self._interface_datatype):
            if item._component is None:
                item._component = weakref_ref(self)
                if hasattr(self, "_active"):
                    self._active |= getattr(item, '_active', True)
                self._data.insert(i, item)
                item._index = i
                return
            # see note about allowing components to live in more than
            # one container
            raise ValueError(
                "Invalid component object assignment to ComponentList "
                "%s at index %s. A parent component has already been "
                "assigned the object: %s" % (self.name, i, item.parent_component().name)
            )
        # see note about implicit assignment and update
        raise TypeError(
            "ComponentList must be assigned objects "
            "of type %s. Invalid type for key %s: %s"
            % (self._interface_datatype.__name__, i, type(item))
        )

    # Since we don't currently allow objects to be assigned when their
    # parent component is already set, it would make sense to reset
    # the parent component back to None for an object being removed
    # from this container; thus, allowing objects to be naturally
    # transferred between containers if one so chooses. This would be
    # more complicated if allowing components to live in more than one
    # container (see note above __setitem__), and maybe this is the
    # reason not to support that.
    def __delitem__(self, i):
        obj = self._data[i]
        obj._component = None
        del self._data[i]

    def __getitem__(self, i):
        return self._data[i]

    def __len__(self):
        return self._data.__len__()

    #
    # Override a few default implementations on MutableSequence
    #

    # We want to avoid generating Pyomo expressions by comparing values
    def __contains__(self, item):
        item_id = id(item)
        return any(item_id == id(_v) for _v in self._data)

    # We want to avoid generating Pyomo expressions by comparing values
    def index(self, item, start=0, stop=None):
        '''S.index(value, [start, [stop]]) -> integer -- return first index of value.

        Raises ValueError if the value is not present.
        '''
        if start is not None and start < 0:
            start = max(len(self) + start, 0)
        if stop is not None and stop < 0:
            stop += len(self)

        item_id = id(item)
        i = start
        while stop is None or i < stop:
            try:
                if id(self[i]) == item_id:
                    return i
            except IndexError:
                break
            i += 1
        raise ValueError

    # We want to avoid generating Pyomo expressions by comparing values
    def count(self, item):
        'S.count(value) -> integer -- return number of occurrences of value'
        item_id = id(item)
        cnt = sum(1 for _v in self._data if id(_v) == item_id)
        assert cnt == 1
        return cnt

    # Avoid errors related to calling __setitem__
    # with a component that is already owned
    def reverse(self):
        'S.reverse() -- reverse *IN PLACE*'
        n = len(self)
        data = self._data
        for i in range(n // 2):
            data[i], data[n - i - 1] = data[n - i - 1], data[i]


#
# ComponentList needs to come before IndexedComponent
# (or subclasses of) so we can override certain methods
#


class XVarList(ComponentList, IndexedVar):
    def __init__(self, *args, **kwds):
        IndexedVar.__init__(self, Any, **kwds)
        # Constructor for ComponentList needs to
        # go last in order to handle any initialization
        # iterable as an argument
        ComponentList.__init__(self, VarData, *args, **kwds)


class XConstraintList(ComponentList, IndexedConstraint):
    def __init__(self, *args, **kwds):
        IndexedConstraint.__init__(self, Any, **kwds)
        # Constructor for ComponentList needs to
        # go last in order to handle any initialization
        # iterable as an argument
        ComponentList.__init__(self, ConstraintData, *args, **kwds)


class XObjectiveList(ComponentList, IndexedObjective):
    def __init__(self, *args, **kwds):
        IndexedObjective.__init__(self, Any, **kwds)
        # Constructor for ComponentList needs to
        # go last in order to handle any initialization
        # iterable as an argument
        ComponentList.__init__(self, ObjectiveData, *args, **kwds)


class XExpressionList(ComponentList, IndexedExpression):
    def __init__(self, *args, **kwds):
        IndexedExpression.__init__(self, Any, **kwds)
        # Constructor for ComponentList needs to
        # go last in order to handle any initialization
        # iterable as an argument
        ComponentList.__init__(self, ExpressionData, *args, **kwds)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentSet
from pyomo.core.base.indexed_component import normalize_index
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice
from pyomo.core.base.global_set import UnindexedComponent_set


def _to_iterable(source):
    iterable_scalars = (str, bytes)
    if hasattr(source, '__iter__'):
        if isinstance(source, iterable_scalars):
            yield source
        else:
            for obj in source:
                yield obj
    else:
        yield source


def get_component_call_stack(comp, context=None):
    """Get the call stack necessary to locate a `Component`

    The call stack is a `list` of `tuple`s where the first entry is a
    code for `__getattr__` or  `__getitem__`, using the same convention
    as `IndexedComponent_slice`. The second entry is the argument of
    the corresponding function. Following this sequence of calls from
    `context` (or the top-level model if `context is None`) will
    produce comp.

    Parameters:
    -----------
    comp : `pyomo.core.base.component.Component`
        The component to locate
    context : `pyomo.core.base.block.Block`
        The block within which to locate the component. If `None`, the
        top-level model will be used.

    Returns:
    --------
    `list` : Contains the necessary method calls and their arguments.
             Note that the calls should be applied in reverse order.
             This is the opposite direction as in IndexedComponent_slice.

    """
    # If comp is context, an empty call stack is returned.
    # A component is not said to exist in "the context of" itself.
    call_stack = []
    while comp.parent_block() is not None:
        # If parent_block is None, comp is the root
        # of the model, so we don't add anything else
        # to the call stack.
        if comp is context:
            # We are done
            break

        parent_component = comp.parent_component()
        # Add (get_item, index) to the call stack
        if parent_component.is_indexed() and parent_component is not comp:
            # I.e. `comp` is a data object
            call_stack.append((IndexedComponent_slice.get_item, comp.index()))

        if parent_component is context:
            # We are done
            break

        # Add (get_attribute, name) to the call stack
        call_stack.append(
            (IndexedComponent_slice.get_attribute, parent_component.local_name)
        )
        comp = comp.parent_block()
    return call_stack


def slice_component_along_sets(comp, sets, context=None):
    """Slice a component along the indices corresponding to some sets,
    wherever they appear in the component's block hierarchy.

    Given a component or component data object, for all parent components
    and parent blocks between the object and the `context` block, replace
    any index corresponding to a set in `sets` with slices or an
    ellipsis.

    Parameters:
    -----------
    comp: :class:`Component` or :class:`ComponentData`
        Component whose parent structure to search and replace
    sets: `~pyomo.common.collections.ComponentSet`
        Contains the sets to replace with slices
    context: :class:`Block` or :class:`BlockData`
        Block below which to search for sets

    Returns:
    --------
    `~pyomo.core.base.indexed_component_slice.IndexedComponent_slice`:
        Slice of `comp` with wildcards replacing the indices of `sets`

    """
    # Cast to ComponentSet so a tuple or list of sets is an appropriate
    # argument. Otherwise a tuple or list of sets will potentially
    # silently do the wrong thing (as inclusion in a tuple or list does
    # not distinguish between different sets with the same elements).
    sets = ComponentSet(sets)

    if context is None:
        context = comp.model()
    call_stack = get_component_call_stack(comp, context)

    # Maintain a pointer to the component so we can get
    # the index set and know which locations to slice.
    comp = context
    sliced_comp = context
    while call_stack:
        call, arg = call_stack.pop()
        if call is IndexedComponent_slice.get_attribute:
            comp = getattr(comp, arg)
            sliced_comp = getattr(sliced_comp, arg)
        elif call is IndexedComponent_slice.get_item:
            index_set = comp.index_set()
            comp = comp[arg]
            # Process arg to replace desired indices with slices.
            location_set_map = get_location_set_map(arg, index_set)
            arg = replace_indices(arg, location_set_map, sets)
            if normalize_index.flatten:
                arg = normalize_index(arg)
            sliced_comp = sliced_comp[arg]

    return sliced_comp


def replace_indices(index, location_set_map, sets):
    """Use `location_set_map` to replace values in `index` with slices
    or an Ellipsis.

    Parameters:
    -----------
    index: `tuple` or scalar
        Index whose values to replace
    location_set_map: `dict`
        Maps locations ("indices") within the index to their
        corresponding set
    sets: `pyomo.common.collections.ComponentSet`
        Contains the sets to replace with slices

    Returns:
    --------
    `tuple`: Index with values replaced by slices

    """
    sets = ComponentSet(sets)
    index = tuple(_to_iterable(index))
    new_index = []
    loc = 0
    len_index = len(index)
    while loc < len_index:
        val = index[loc]
        _set = location_set_map[loc]
        dimen = _set.dimen
        if _set not in sets:
            new_index.append(val)
        elif dimen is not None:
            new_index.append(slice(None, None, None))
        else:
            dimen_none_set = _set
            new_index.append(Ellipsis)
            loc += 1
            while loc < len_index:
                # Skip all adjacent locations belonging to the same
                # set. These are covered by the Ellipsis.
                _set = location_set_map[loc]
                if _set is not dimen_none_set:
                    break
                loc += 1
            continue
        loc += 1
    return tuple(new_index)


def get_location_set_map(index, index_set):
    """Map each value in an index to the set from which it originates

    This function iterates over the "subsets" of `index_set` in the
    forward direction, assigning sets to each value in `index`, until
    it finds a set of dimension `None`. Then it iterates over the
    reversed list of subsets, assigning sets, until it encounters
    the same set of dimension `None`. All remaining values are assigned
    to the set of dimension `None`. If a second such set is found,
    an error is raised.

    Parameters:
    -----------
    index: `tuple` or hashable scalar
        The index whose values will be assigned to sets
    index_set: `pyomo.core.base.set.SetProduct` or `pyomo.core.base.set.Set`
        The index set from which `index` originates

    Returns:
    --------
    `dict`: Maps the "locations" (indices) within the index to the set
        from which it originates.

    """
    index = tuple(_to_iterable(index))
    len_index = len(index)
    locations_left = set(range(len_index))
    location_set_map = {}

    # Determine if we should exit early
    if index_set is UnindexedComponent_set:
        # index == (None,) and dimen is None in this case,
        # so without this catch, the function would return
        # {0: None}, which is not what we want for an
        # unindexed component.
        return {0: UnindexedComponent_set}
    elif not normalize_index.flatten:
        raise ValueError(
            'get_location_set_map does not support the case where '
            'normalize_index.flatten is False.'
            # Although in this case, the location of an index should
            # just be its position in the subsets list, so maybe
            # the info we need is actually more simple to obtain.
        )

    subsets = list(index_set.subsets())

    n_subsets = len(subsets)

    location = 0
    dimen_none_set = None
    dimen_none_set_coord = None
    # Step through subsets in forward order to assign as many
    # locations as possible.
    for sub_coord, sub in enumerate(subsets):
        dimen = sub.dimen
        if dimen is None:
            dimen_none_set = sub
            dimen_none_set_coord = sub_coord
            break
        for i in range(dimen):
            location_set_map[location + i] = sub
            locations_left.remove(location + i)
        location += dimen

    # We are either done or have encountered a set of dimen None
    if not locations_left:
        return location_set_map

    location = len_index - 1
    # Step through subsets in reverse order, assigning locations,
    # until a set of dimen==None is encountered.
    for sub_coord, sub in enumerate(reversed(subsets)):
        sub_coord = n_subsets - sub_coord - 1
        dimen = sub.dimen
        if dimen is None:
            if dimen_none_set_coord != sub_coord:
                # Make sure this set is the same one we encountered
                # earlier. It is sufficient to check the coordinate.
                raise RuntimeError(
                    'Cannot get locations when multiple sets of dimen==None '
                    'are present.'
                    '\nFound %s at position %s and %s at position %s.'
                    '\nLocation is ambiguous in this case.'
                    % (dimen_none_set, dimen_none_set_coord, sub, sub_coord)
                )
            break
        for i in range(dimen):
            location_set_map[location - i] = sub
            locations_left.remove(location - i)
        location -= sub.dimen

    for loc in locations_left:
        # All remaining locations, that cannot be accessed from some
        # constant offset from the beginning or end of the tuple,
        # must belong to the dimen-None set.
        location_set_map[loc] = dimen_none_set

    return location_set_map



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.opt import check_optimal_termination
from pyomo.common.dependencies import attempt_import

np, numpy_available = attempt_import(
    'numpy', 'inverse_reduced_hessian numpy', minimum_version='1.13.0'
)
scipy, scipy_available = attempt_import(
    'scipy', 'inverse_reduced_hessian requires scipy'
)

if numpy_available:
    from pyomo.contrib.pynumero.asl import AmplInterface

    asl_available = AmplInterface.available()
else:
    asl_available = False

if not (numpy_available and scipy_available and asl_available):
    raise unittest.SkipTest(
        'inverse_reduced_hessian tests require numpy, scipy, and asl'
    )
from pyomo.common.dependencies import pandas as pd, pandas_available

ipopt_solver = pyo.SolverFactory('ipopt')
if not ipopt_solver.available(exception_flag=False):
    raise unittest.SkipTest('ipopt is not available')

numdiff_available = True
try:
    import numdifftools as nd
except:
    numdiff_available = False

from pyomo.contrib.interior_point.inverse_reduced_hessian import (
    inv_reduced_hessian_barrier,
)


class TestInverseReducedHessian(unittest.TestCase):
    # the original test
    def test_invrh_zavala_thesis(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3])
        m.obj = pyo.Objective(
            expr=(m.x[1] - 1) ** 2 + (m.x[2] - 2) ** 2 + (m.x[3] - 3) ** 2
        )
        m.c1 = pyo.Constraint(expr=m.x[1] + 2 * m.x[2] + 3 * m.x[3] == 0)

        status, invrh = inv_reduced_hessian_barrier(m, [m.x[2], m.x[3]])
        expected_invrh = np.asarray(
            [[0.35714286, -0.21428571], [-0.21428571, 0.17857143]]
        )
        np.testing.assert_array_almost_equal(invrh, expected_invrh)

    # test by DLW, April 2020
    def _simple_model(self, add_constraint=False):
        # Hardwired to have two x columns and one y
        # if add_constraint is true, there is a binding constraint on b0
        data = pd.DataFrame(
            [
                [1, 1.1, 0.365759306],
                [2, 1.2, 4],
                [3, 1.3, 4.8876684],
                [4, 1.4, 5.173455561],
                [5, 1.5, 2.093799081],
                [6, 1.6, 9],
                [7, 1.7, 6.475045106],
                [8, 1.8, 8.127111268],
                [9, 1.9, 6],
                [10, 1.21, 10.20642714],
                [11, 1.22, 13.08211636],
                [12, 1.23, 10],
                [13, 1.24, 15.38766047],
                [14, 1.25, 14.6587746],
                [15, 1.26, 13.68608604],
                [16, 1.27, 14.70707893],
                [17, 1.28, 18.46192779],
                [18, 1.29, 15.60649164],
            ],
            columns=['tofu', 'chard', 'y'],
        )

        model = pyo.ConcreteModel()

        model.b0 = pyo.Var(initialize=0)
        model.bindexes = pyo.Set(initialize=['tofu', 'chard'])
        model.b = pyo.Var(model.bindexes, initialize=1)

        # try to make trouble
        if add_constraint:
            model.binding_constraint = pyo.Constraint(expr=model.b0 >= 10)

        # The columns need to have unique values (or you get warnings)
        def response_rule(m, t, c):
            expr = m.b0 + m.b['tofu'] * t + m.b['chard'] * c
            return expr

        model.response_function = pyo.Expression(
            data.tofu, data.chard, rule=response_rule
        )

        def SSE_rule(m):
            return sum(
                (data.y[i] - m.response_function[data.tofu[i], data.chard[i]]) ** 2
                for i in data.index
            )

        model.SSE = pyo.Objective(rule=SSE_rule, sense=pyo.minimize)

        return model

    @unittest.skipIf(not numdiff_available, "numdifftools missing")
    @unittest.skipIf(not pandas_available, "pandas missing")
    def test_3x3_using_linear_regression(self):
        """simple linear regression with two x columns, so 3x3 Hessian"""

        model = self._simple_model()
        solver = pyo.SolverFactory("ipopt")
        status = solver.solve(model)
        self.assertTrue(check_optimal_termination(status))
        tstar = [
            pyo.value(model.b0),
            pyo.value(model.b['tofu']),
            pyo.value(model.b['chard']),
        ]

        def _ndwrap(x):
            # wrapper for numdiff call
            model.b0.fix(x[0])
            model.b["tofu"].fix(x[1])
            model.b["chard"].fix(x[2])
            rval = pyo.value(model.SSE)
            return rval

        H = nd.Hessian(_ndwrap)(tstar)
        HInv = np.linalg.inv(H)

        model.b0.fixed = False
        model.b["tofu"].fixed = False
        model.b["chard"].fixed = False
        status, H_inv_red_hess = inv_reduced_hessian_barrier(
            model, [model.b0, model.b["tofu"], model.b["chard"]]
        )
        # this passes at decimal=6, BTW
        np.testing.assert_array_almost_equal(HInv, H_inv_red_hess, decimal=3)

    @unittest.skipIf(not numdiff_available, "numdifftools missing")
    @unittest.skipIf(not pandas_available, "pandas missing")
    def test_with_binding_constraint(self):
        """there is a binding constraint"""

        model = self._simple_model(add_constraint=True)

        status, H_inv_red_hess = inv_reduced_hessian_barrier(
            model, [model.b0, model.b["tofu"], model.b["chard"]]
        )
        print("test_with_binding_constraint should see an error raised.")


if __name__ == '__main__':
    unittest.main()
